# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Export a YOLOv5 PyTorch model to TorchScript, ONNX, CoreML, TensorFlow (saved_model, pb, TFLite, TF.js,) formats
TensorFlow exports authored by https://github.com/zldrobit

Usage:
    $ python path/to/export.py --weights yolov5s.pt --include torchscript onnx coreml saved_model pb tflite tfjs

Inference:
    $ python path/to/detect.py --weights yolov5s.pt
                                         yolov5s.onnx  (must export with --dynamic)
                                         yolov5s_saved_model
                                         yolov5s.pb
                                         yolov5s.tflite

TensorFlow.js:
    $ cd .. && git clone https://github.com/zldrobit/tfjs-yolov5-example.git && cd tfjs-yolov5-example
    $ npm install
    $ ln -s ../../yolov5/yolov5s_web_model public/yolov5s_web_model
    $ npm start
"""

import argparse
import subprocess
import sys
import time
import os
from pathlib import Path

import torch
import torch.nn as nn
from torch.utils.mobile_optimizer import optimize_for_mobile

FILE = Path(__file__).resolve()
ROOT = FILE.parents[0]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH
ROOT = ROOT.relative_to(Path.cwd())  # relative

from models.common import Conv
from models.experimental import attempt_load
from models.yolo import Detect
from utils.activations import SiLU
from utils.datasets import LoadImages
from utils.general import colorstr, check_dataset, check_img_size, check_requirements, file_size, print_args, \
    set_logging, url2file
from utils.torch_utils import select_device


def export_torchscript(model, im, file, optimize, prefix=colorstr('TorchScript:')):
    # YOLOv5 TorchScript model export
    try:
        dev_type = next(model.parameters()).device.type
        print(f'\n{prefix} starting export with torch {torch.__version__}...')
        fn = str(file.name)
        fn = fn.replace('.pt','')
        pth = str(file.parent)
        h,w = im.shape[-2:]
        batch_size = im.shape[0]
        f = os.path.join(pth, '{}_{}_{}_{}_B{}_ts.pt'.format(fn, dev_type, h, w, batch_size))

        ts = torch.jit.trace(model, im, strict=False)
        (optimize_for_mobile(ts) if optimize else ts).save(f)

        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'{prefix} export failure: {e}')


def export_onnx(model, im, file, opset, train, dynamic, simplify, prefix=colorstr('ONNX:')):
    # YOLOv5 ONNX export
    try:
        check_requirements(('onnx',))
        import onnx

        print(f'\n{prefix} starting export with onnx {onnx.__version__}...')
        f = file.with_suffix('.onnx')

        torch.onnx.export(model, im, f, verbose=False, opset_version=opset,
                          training=torch.onnx.TrainingMode.TRAINING if train else torch.onnx.TrainingMode.EVAL,
                          do_constant_folding=not train,
                          input_names=['images'],
                          output_names=['output'],
                          dynamic_axes={'images': {0: 'batch', 2: 'height', 3: 'width'},  # shape(1,3,640,640)
                                        'output': {0: 'batch', 1: 'anchors'}  # shape(1,25200,85)
                                        } if dynamic else None)

        # Checks
        model_onnx = onnx.load(f)  # load onnx model
        onnx.checker.check_model(model_onnx)  # check onnx model
        # print(onnx.helper.printable_graph(model_onnx.graph))  # print

        # Simplify
        if simplify:
            try:
                check_requirements(('onnx-simplifier',))
                import onnxsim

                print(f'{prefix} simplifying with onnx-simplifier {onnxsim.__version__}...')
                model_onnx, check = onnxsim.simplify(
                    model_onnx,
                    dynamic_input_shape=dynamic,
                    input_shapes={'images': list(im.shape)} if dynamic else None)
                assert check, 'assert check failed'
                onnx.save(model_onnx, f)
            except Exception as e:
                print(f'{prefix} simplifier failure: {e}')
        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
        print(f"{prefix} run --dynamic ONNX model inference with: 'python detect.py --weights {f}'")
    except Exception as e:
        print(f'{prefix} export failure: {e}')


def export_coreml(model, im, file, prefix=colorstr('CoreML:')):
    # YOLOv5 CoreML export
    ct_model = None
    try:
        check_requirements(('coremltools',))
        import coremltools as ct

        print(f'\n{prefix} starting export with coremltools {ct.__version__}...')
        f = file.with_suffix('.mlmodel')

        model.train()  # CoreML exports should be placed in model.train() mode
        ts = torch.jit.trace(model, im, strict=False)  # TorchScript model
        ct_model = ct.convert(ts, inputs=[ct.ImageType('image', shape=im.shape, scale=1 / 255.0, bias=[0, 0, 0])])
        ct_model.save(f)

        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'\n{prefix} export failure: {e}')

    return ct_model


def export_saved_model(model, im, file, dynamic,
                       tf_nms=False, agnostic_nms=False, topk_per_class=100, topk_all=100, iou_thres=0.45,
                       conf_thres=0.25, prefix=colorstr('TensorFlow saved_model:')):
    # YOLOv5 TensorFlow saved_model export
    keras_model = None
    try:
        import tensorflow as tf
        from tensorflow import keras
        from models.tf import TFModel, TFDetect

        print(f'\n{prefix} starting export with tensorflow {tf.__version__}...')
        f = str(file).replace('.pt', '_saved_model')
        batch_size, ch, *imgsz = list(im.shape)  # BCHW

        tf_model = TFModel(cfg=model.yaml, model=model, nc=model.nc, imgsz=imgsz)
        im = tf.zeros((batch_size, *imgsz, 3))  # BHWC order for TensorFlow
        y = tf_model.predict(im, tf_nms, agnostic_nms, topk_per_class, topk_all, iou_thres, conf_thres)
        inputs = keras.Input(shape=(*imgsz, 3), batch_size=None if dynamic else batch_size)
        outputs = tf_model.predict(inputs, tf_nms, agnostic_nms, topk_per_class, topk_all, iou_thres, conf_thres)
        keras_model = keras.Model(inputs=inputs, outputs=outputs)
        keras_model.trainable = False
        keras_model.summary()
        keras_model.save(f, save_format='tf')

        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'\n{prefix} export failure: {e}')

    return keras_model


def export_pb(keras_model, im, file, prefix=colorstr('TensorFlow GraphDef:')):
    # YOLOv5 TensorFlow GraphDef *.pb export https://github.com/leimao/Frozen_Graph_TensorFlow
    try:
        import tensorflow as tf
        from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2

        print(f'\n{prefix} starting export with tensorflow {tf.__version__}...')
        f = file.with_suffix('.pb')

        m = tf.function(lambda x: keras_model(x))  # full model
        m = m.get_concrete_function(tf.TensorSpec(keras_model.inputs[0].shape, keras_model.inputs[0].dtype))
        frozen_func = convert_variables_to_constants_v2(m)
        frozen_func.graph.as_graph_def()
        tf.io.write_graph(graph_or_graph_def=frozen_func.graph, logdir=str(f.parent), name=f.name, as_text=False)

        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'\n{prefix} export failure: {e}')


def export_tflite(keras_model, im, file, int8, data, ncalib, prefix=colorstr('TensorFlow Lite:')):
    # YOLOv5 TensorFlow Lite export
    try:
        import tensorflow as tf
        from models.tf import representative_dataset_gen

        print(f'\n{prefix} starting export with tensorflow {tf.__version__}...')
        batch_size, ch, *imgsz = list(im.shape)  # BCHW
        f = str(file).replace('.pt', '-fp16.tflite')

        converter = tf.lite.TFLiteConverter.from_keras_model(keras_model)
        converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS]
        converter.target_spec.supported_types = [tf.float16]
        converter.optimizations = [tf.lite.Optimize.DEFAULT]
        if int8:
            dataset = LoadImages(check_dataset(data)['train'], img_size=imgsz, auto=False)  # representative data
            converter.representative_dataset = lambda: representative_dataset_gen(dataset, ncalib)
            converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]
            converter.target_spec.supported_types = []
            converter.inference_input_type = tf.uint8  # or tf.int8
            converter.inference_output_type = tf.uint8  # or tf.int8
            converter.experimental_new_quantizer = False
            f = str(file).replace('.pt', '-int8.tflite')

        tflite_model = converter.convert()
        open(f, "wb").write(tflite_model)
        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')

    except Exception as e:
        print(f'\n{prefix} export failure: {e}')


def export_tfjs(keras_model, im, file, prefix=colorstr('TensorFlow.js:')):
    # YOLOv5 TensorFlow.js export
    try:
        check_requirements(('tensorflowjs',))
        import re
        import tensorflowjs as tfjs

        print(f'\n{prefix} starting export with tensorflowjs {tfjs.__version__}...')
        f = str(file).replace('.pt', '_web_model')  # js dir
        f_pb = file.with_suffix('.pb')  # *.pb path
        f_json = f + '/model.json'  # *.json path

        cmd = f"tensorflowjs_converter --input_format=tf_frozen_model " \
              f"--output_node_names='Identity,Identity_1,Identity_2,Identity_3' {f_pb} {f}"
        subprocess.run(cmd, shell=True)

        json = open(f_json).read()
        with open(f_json, 'w') as j:  # sort JSON Identity_* in ascending order
            subst = re.sub(
                r'{"outputs": {"Identity.?.?": {"name": "Identity.?.?"}, '
                r'"Identity.?.?": {"name": "Identity.?.?"}, '
                r'"Identity.?.?": {"name": "Identity.?.?"}, '
                r'"Identity.?.?": {"name": "Identity.?.?"}}}',
                r'{"outputs": {"Identity": {"name": "Identity"}, '
                r'"Identity_1": {"name": "Identity_1"}, '
                r'"Identity_2": {"name": "Identity_2"}, '
                r'"Identity_3": {"name": "Identity_3"}}}',
                json)
            j.write(subst)

        print(f'{prefix} export success, saved as {f} ({file_size(f):.1f} MB)')
    except Exception as e:
        print(f'\n{prefix} export failure: {e}')


@torch.no_grad()
def run(data=ROOT / 'data/coco128.yaml',  # 'dataset.yaml path'
        weights=ROOT / 'yolov5s.pt',  # weights path
        imgsz=(640, 640),  # image (height, width)
        batch_size=1,  # batch size
        device='cpu',  # cuda device, i.e. 0 or 0,1,2,3 or cpu
        include=('torchscript', 'onnx', 'coreml'),  # include formats
        half=False,  # FP16 half-precision export
        inplace=False,  # set YOLOv5 Detect() inplace=True
        train=False,  # model.train() mode
        optimize=False,  # TorchScript: optimize for mobile
        int8=False,  # CoreML/TF INT8 quantization
        dynamic=False,  # ONNX/TF: dynamic axes
        simplify=False,  # ONNX: simplify model
        opset=12,  # ONNX: opset version
        topk_per_class=100,  # TF.js NMS: topk per class to keep
        topk_all=100,  # TF.js NMS: topk for all classes to keep
        iou_thres=0.45,  # TF.js NMS: IoU threshold
        conf_thres=0.25  # TF.js NMS: confidence threshold
        ):
    t = time.time()
    include = [x.lower() for x in include]
    tf_exports = list(x in include for x in ('saved_model', 'pb', 'tflite', 'tfjs'))  # TensorFlow exports
    imgsz *= 2 if len(imgsz) == 1 else 1  # expand
    file = Path(url2file(weights) if str(weights).startswith(('http:/', 'https:/')) else weights)

    # Load PyTorch model
    device = select_device(device)
    assert not (device.type == 'cpu' and half), '--half only compatible with GPU export, i.e. use --device 0'
    model = attempt_load(weights, map_location=device, inplace=True, fuse=True)  # load FP32 model
    nc, names = model.nc, model.names  # number of classes, class names

    # Input
    gs = int(max(model.stride))  # grid size (max stride)
    imgsz = [check_img_size(x, gs) for x in imgsz]  # verify img_size are gs-multiples
    im = torch.zeros(batch_size, 3, *imgsz).to(device)  # image size(1,3,320,192) BCHW iDetection

    # Update model
    if half:
        im, model = im.half(), model.half()  # to FP16
    model.train() if train else model.eval()  # training mode = no Detect() layer grid construction
    for k, m in model.named_modules():
        if isinstance(m, Conv):  # assign export-friendly activations
            if isinstance(m.act, nn.SiLU):
                m.act = SiLU()
        elif isinstance(m, Detect):
            m.inplace = inplace
            m.onnx_dynamic = dynamic
            # m.forward = m.forward_export  # assign forward (optional)

    for _ in range(2):
        y = model(im)  # dry runs
    print(f"\n{colorstr('PyTorch:')} starting from {file} ({file_size(file):.1f} MB)")

    # Exports
    if 'torchscript' in include:
        export_torchscript(model, im, file, optimize)
    if 'onnx' in include:
        export_onnx(model, im, file, opset, train, dynamic, simplify)
    if 'coreml' in include:
        export_coreml(model, im, file)

    # TensorFlow Exports
    if any(tf_exports):
        pb, tflite, tfjs = tf_exports[1:]
        assert not (tflite and tfjs), 'TFLite and TF.js models must be exported separately, please pass only one type.'
        model = export_saved_model(model, im, file, dynamic, tf_nms=tfjs, agnostic_nms=tfjs,
                                   topk_per_class=topk_per_class, topk_all=topk_all, conf_thres=conf_thres,
                                   iou_thres=iou_thres)  # keras model
        if pb or tfjs:  # pb prerequisite to tfjs
            export_pb(model, im, file)
        if tflite:
            export_tflite(model, im, file, int8=int8, data=data, ncalib=100)
        if tfjs:
            export_tfjs(model, im, file)

    # Finish
    print(f'\nExport complete ({time.time() - t:.2f}s)'
          f"\nResults saved to {colorstr('bold', file.parent.resolve())}"
          f'\nVisualize with https://netron.app')


def parse_opt():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data', type=str, default=ROOT / 'data/coco128.yaml', help='dataset.yaml path')
    parser.add_argument('--weights', type=str, default=ROOT / 'yolov5l.pt', help='weights path')
    parser.add_argument('--imgsz', '--img', '--img-size', nargs='+', type=int, default=[640, 640], help='image (h, w)')
    parser.add_argument('--batch-size', type=int, default=1, help='batch size')
    parser.add_argument('--device', default='0', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
    parser.add_argument('--half', action='store_true', help='FP16 half-precision export')
    parser.add_argument('--inplace', action='store_true', help='set YOLOv5 Detect() inplace=True')
    parser.add_argument('--train', action='store_true', help='model.train() mode')
    parser.add_argument('--optimize', action='store_true', help='TorchScript: optimize for mobile')
    parser.add_argument('--int8', action='store_true', help='CoreML/TF INT8 quantization')
    parser.add_argument('--dynamic', action='store_true', help='ONNX/TF: dynamic axes')
    parser.add_argument('--simplify', action='store_true', help='ONNX: simplify model')
    parser.add_argument('--opset', type=int, default=13, help='ONNX: opset version')
    parser.add_argument('--topk-per-class', type=int, default=100, help='TF.js NMS: topk per class to keep')
    parser.add_argument('--topk-all', type=int, default=100, help='TF.js NMS: topk for all classes to keep')
    parser.add_argument('--iou-thres', type=float, default=0.45, help='TF.js NMS: IoU threshold')
    parser.add_argument('--conf-thres', type=float, default=0.25, help='TF.js NMS: confidence threshold')
    parser.add_argument('--include', nargs='+',
                        default=['torchscript'],
                        help='available formats are (torchscript, onnx, coreml, saved_model, pb, tflite, tfjs)')
    opt = parser.parse_args()
    print_args(FILE.stem, opt)
    return opt


def main(opt):
    set_logging()
    run(**vars(opt))


if __name__ == "__main__":
    opt = parse_opt()
    main(opt)



import redis
import json

from .. import Logger

class PublisherWrapper:
  def __init__(self,
               log : Logger,
               publisher=None):
    self.log = log
    self.publisher = publisher

  def maybe_publish(self, channel, eid, dct_message, verbose=2):
    if self.publisher is not None:
      message = json.dumps({eid : dct_message})
      self.publisher.publish(channel, message)
      if verbose >= 1:
        self.log.P("Published 1 message through redis channel '{}'".format(channel), color='g')
      if verbose >= 2:
        self.log.P("  Message: {}".format(message))
    return

class Publisher(object):
  def __init__(self, host=None, port=None, db=None, password=None):
    """
    Parameters:
    ------------
    host: str, optional
      the redis server host
      Default to None ('localhost')

    port: int, optional
      the redis server port
      Default to None (6379)

    db: int, optional
      the redis server db
      Default to None (0)

    password: str, optional
      The password of the redis server
      Default to None - if the server does not have any password
    """

    if host is None:
      host = 'localhost'

    if port is None:
      port = 6379

    if db is None:
      db = 0

    self._publisher = redis.Redis(
      host=host, port=port, db=db,
      password=password,
      socket_keepalive=True,
      health_check_interval=10
    )
    return
  
  def publish(self, channel, message):
    """
    Publishes a message through a redis channel

    Parameters:
    -----------
    channel: str, mandatory
      The redis channel through which the message is sent

    message: str, mandatory
      The message to send
    """
    
    # Reset the connection pool everytime to avoid potential timeouts
    self._publisher.connection_pool.reset()
    return self._publisher.publish(channel, message)
    



import cv2
import os

from naeural_core import DecentrAIObject
from naeural_core import Logger

class VideoRecorder(DecentrAIObject):
  """
  Class that will be initialized in specific plugins implementations (not base plugin!) for recording live input (frames)
  that causes errors in a stateful plugin.

  Steps for using:
  1. initilize
  2. call `maybe_start` and `write` whenever the input data comes
  3. call `stop` when the recording is done (in the most cases, when an Exception occurs)
  """

  def __init__(self, log : Logger, name, **kwargs):
    self._recording = None
    self._name = name

    self.__force_stopped = False
    super(VideoRecorder, self).__init__(log=log, prefix_log='[VREC]', **kwargs)
    return

  def startup(self):
    super().startup()
    return

  def maybe_start(self, height, width):
    if self._recording is not None:
      return

    fld = os.path.join(
      self.log.get_output_folder(),
      'recordings'
    )
    if not os.path.exists(fld):
      os.makedirs(fld)

    path = os.path.join(
      fld,
      '{}_{}.avi'.format(self.log.now_str(), self._name)
    )

    fourcc = cv2.VideoWriter_fourcc(*'XVID')
    fps = 25
    out = cv2.VideoWriter(path, fourcc, fps, (width, height))
    self._recording = {
      'CODEC': out,
      'PATH': path
    }
    self.P('Started recording {} in {}'.format(self._name, path))
    return

  def stop(self):
    if self.__force_stopped:
      self.P("Recording {} was already force stopped".format(self._name))
      return

    self._recording['CODEC'].release()
    self.P('Stopped recording {}. Video: {}'.format(
      self._name,
      self._recording['PATH']
    ))

    return

  def write(self, frame):
    self._recording['CODEC'].write(frame)

    size_gb = os.path.getsize(self._recording['PATH']) / (2**30)
    thr_gb = 10.0

    if size_gb >= thr_gb:
      self.P("Force stop the recording as it occupies more than {}GB".format(thr_gb))
      self.stop()
      self.__force_stopped = True
    #endif
    return






import traceback
from naeural_core import Logger
import os

class MSCT:
  NO_STARTUP_WAIT = 'NO_STARTUP_WAIT'
  CONFIG_ENDPOINTS = 'CONFIG_ENDPOINTS'
  
  SIGNATURE = 'SIGNATURE'
  
  PROCESS = 'PROCESS'
  DISABLED = 'DISABLED'
  DATA = 'data'
  VER = 'core_ver'
  GW_UPTIME = 'gw-uptime'
  TIME = 'time'
  SERVER_CLASS = 'SERVER_CLASS'
  HOST = 'HOST'
  DESCRIPTION = 'DESCRIPTION'
  NR_WORKERS = 'NR_WORKERS'
  PATHS = 'PATHS'
  ONLINE = 'ONLINE'
  PORT = 'PORT'
  ERROR = 'ERROR'
  URLS = 'URLS'
  UPTIME = 'UPTIME'
  START = 'START'
  SYSTEM_STATUS = 'SYSTEM_STATUS'
  DEFAULT_SERVER = 'DEFAULT_SERVER'
  AVAIL_SERVERS = 'AVAIL_SERVERS'
  
  NOTIF_NOTIFICATION_TYPE = 'NOTIFICATION_TYPE'
  NOTIF_MODULE = 'MODULE'
  NOTIF_TIME = 'TIMESTAMP'
  
  DOWNLOAD_FILE_COMMAND = 'DOWNLOAD'  
  DOWNLOAD_FILE_PATH = 'DOWNLOAD_FILE_PATH'
  
  RULE_LIST = '/list_servers'
  RULE_SHUTDOWN = '/shutdown'
  RULE_DEFAULT = '/analyze'
  RULE_START = '/start_server'
  RULE_KILL = '/kill_server'
  RULE_SYS = '/system_status'
  
  RULE_NOTIF = '/notifications'
  RULE_RUN = '/run'
  RULE_UPDATE_WORKERS = '/update_workers'
  RULE_PATHS = '/get_paths'
  
  KILL_CMD = 'SAFE_KILL_SERVER_CMD' # SAFEWEB_AI_KILL_SERVER

def get_api_request_body(request, log : Logger, sender=None):
  try:
    method = request.method
    args_data = request.args
    form_data = request.form
    json_data = request.json

    if method == 'GET':
      # parameters in URL
      base_params = args_data
    else:
      # parameters in form
      base_params = form_data
      if len(base_params) == 0:
        # params in json?
        base_params = json_data
    #endif

    if base_params is not None:
      params = dict(base_params)
    else:
      params = {}
    #endif
  except Exception as e:
    s = 'sender={}\n\ntraceback={}\n\n\n\nrequest.data={}'.format(sender, traceback.format_exc(), str(request.data))
    fn = 'error_{}'.format(log.now_str())
    if sender is not None:
      fn += '_{}'.format(sender)
    with open(os.path.join(log.get_output_folder(), '{}.txt'.format(fn)), 'wt') as fh:
      fh.write(s)

  return params



"""
Module: centroid_object_tracker.py

This module provides the CentroidObjectTracker class, which implements object tracking using 
centroid-based linear tracking and the SORT algorithm.

IMPORTANT / WARNING: 
	This class should be modified/tuned within individual implementations of complex ecosystems. It is recommended to
	create a `extensions.utils.centroid_object_tracker` unit that will automaticaly replace the current file

The tracker can be configured to use linear tracking, SORT tracking, or both for debugging purposes. 
It maintains the history of detected objects, updates their positions, and handles object registration 
and deregistration based on specified criteria.

"""

import numpy as np
from scipy.spatial import distance as dist
from collections import OrderedDict, deque
from datetime import datetime

from naeural_core.utils.sort import Sort
from naeural_core import constants as ct
from decentra_vision import geometry_methods as gmt


class CentroidObjectTracker:
    def __init__(
        self,
        object_tracking_mode=0,
        linear_max_age=4,
        linear_max_distance=300,
        linear_max_relative_distance=1.2,
        max_dist_scale=1.4,
        center_dist_weight=1,
        hw_dist_weight=0.8,
        sort_min_hits=1,
        sort_max_age=3,
        sort_min_iou=0,
        moved_delta_ratio=0.005,
        linear_reset_minutes=60,
        **kwargs
    ):
        """
        Initialize the CentroidObjectTracker.

        Parameters
        ----------
        object_tracking_mode : int, optional
            Tracking mode. Options are:
            0: Linear tracking (default).
            1: SORT tracking (not recommended).
            2: Both (for debugging purposes only).

        linear_max_age : int, optional
            Maximum number of frames to keep an object without detection before deregistering.

        linear_max_distance : float, optional
            Maximum allowable distance between centroids to consider the same object.

        linear_max_relative_distance : float, optional
            Maximum relative distance allowed between consecutive positions of the same object.
            Calculated as the centroid distance divided by the maximum dimension of the object.

        max_dist_scale : float, optional
            Scale factor for computing maximum distances for distance components.

        center_dist_weight : float, optional
            Weight of the centroid distance in the total distance calculation.

        hw_dist_weight : float, optional
            Weight of the size (height/width) distance in the total distance calculation.

        sort_min_hits : int, optional
            Minimum number of hits to consider a track as valid in SORT tracking.

        sort_max_age : int, optional
            Maximum number of frames to keep a track without detection in SORT tracking.

        sort_min_iou : float, optional
            Minimum IOU for matching in SORT tracking.

        moved_delta_ratio : float, optional
            Not implemented yet.

        linear_reset_minutes : int, optional
            Maximum number of minutes to track an object before resetting.

        **kwargs
            Additional keyword arguments.

        """
        self.nextObjectID = 0
        self.objects = OrderedDict()
        self.disappeared = OrderedDict()
        self.objects_history = OrderedDict()

        self.object_tracking_mode = object_tracking_mode
        self.linear_max_age = linear_max_age
        self.linear_max_distance = linear_max_distance
        self.linear_max_relative_distance = linear_max_relative_distance
        self.linear_reset_minutes = linear_reset_minutes

        self.center_dist_weight = center_dist_weight
        self.hw_dist_weight = hw_dist_weight
        self.max_dist_scale = max_dist_scale

        total_weight = self.hw_dist_weight + self.center_dist_weight
        self.linear_max_distance_center = (
            self.linear_max_distance * self.center_dist_weight / total_weight * self.max_dist_scale
        )
        self.linear_max_distance_hw = (
            self.linear_max_distance * self.hw_dist_weight / total_weight * self.max_dist_scale
        )

        self.sort_min_hits = sort_min_hits
        self.sort_max_age = sort_max_age
        self.sort_min_iou = sort_min_iou

        self.moved_delta_ratio = moved_delta_ratio  # TODO: Implement moved_delta_ratio functionality

        self.sort_tracker = None  # Will be initialized when needed

        # Ensure no kwargs are passed since we don't call super().__init__()
        if kwargs:
            raise ValueError(f"Unexpected keyword arguments: {kwargs}")

    def _maybe_init(self):
        """
        Initialize the SORT tracker if it has not been initialized yet.
        """
        if self.sort_tracker is None:
            self.sort_tracker = Sort(
                min_hits=self.sort_min_hits,
                max_age=self.sort_max_age,
                iou_threshold=self.sort_min_iou
            )

    def register(self, centroid, rectangle):
        """
        Register a new object with a unique ID.

        Parameters
        ----------
        centroid : ndarray
            The centroid of the new object.
        rectangle : list or ndarray
            The bounding box coordinates [startX, startY, endX, endY].

        """
        self.objects[self.nextObjectID] = {
            'appearances': 0,
            'centroid': centroid,
            'rectangle': list(rectangle[:4]),
            'color': ct.RED,
            'first_update': datetime.now(),
            'last_update': datetime.now(),
            'centroid_history': deque([centroid], maxlen=1000),
            'original_position': centroid.copy(),
            'type_history': {'total': 0},
            'type_history_deque': deque(maxlen=100),
            'in_zone_history': deque(maxlen=1000),
            'in_zone_total_seconds': 0
        }
        self.disappeared[self.nextObjectID] = 0
        self.nextObjectID += 1

    def deregister(self, objectID):
        """
        Deregister an object ID by moving its data to history and removing it from tracking.

        Parameters
        ----------
        objectID : int
            The ID of the object to deregister.

        """
        self.objects_history[objectID] = self.objects.pop(objectID)
        self.disappeared.pop(objectID, None)

    def reset_old_objects(self):
        """
        Deregister objects that have been tracked longer than the reset time.

        """
        to_be_removed = []
        current_time = datetime.now()
        for objectID, obj in self.objects.items():
            elapsed_minutes = (current_time - obj['first_update']).total_seconds() / 60
            if elapsed_minutes > self.linear_reset_minutes:
                to_be_removed.append(objectID)

        for objectID in to_be_removed:
            self.deregister(objectID)

    def update_tracker(self, rectangles):
        """
        Update the tracker with the provided rectangles.

        Parameters
        ----------
        rectangles : ndarray
            Array of bounding boxes in the format [startX, startY, endX, endY].

        Returns
        -------
        dict
            Updated tracked objects.

        """
        if self.object_tracking_mode == 0:
            return self.update_linear(rectangles)
        elif self.object_tracking_mode == 1:
            return self.update_sort(rectangles)
        elif self.object_tracking_mode == 2:
            print("WARNING - TRACKING MODE 2 SHOULD BE USED ONLY FOR DEBUGGING PURPOSES")
            linear_results = self.update_linear(rectangles)
            sort_results = self.update_sort(rectangles)
            return {**linear_results, **sort_results}
        else:
            raise NotImplementedError(f"Tracking mode {self.object_tracking_mode} not implemented")

    def update_sort(self, rectangles):
        """
        Update the tracker using SORT algorithm.

        Parameters
        ----------
        rectangles : ndarray
            Array of bounding boxes in the format [startX, startY, endX, endY].

        Returns
        -------
        dict
            Tracked objects using SORT algorithm.

        """
        self._maybe_init()

        if len(rectangles) > 0:
            rectangles_with_confidence = np.hstack((rectangles, np.ones((rectangles.shape[0], 1))))
        else:
            rectangles_with_confidence = np.empty((0, 5))

        tracked_objects = self.sort_tracker.update(rectangles_with_confidence)
        result = {
            f"{int(track[4])}_SORT": {
                'rectangle': [track[0], track[1], track[2], track[3]],
                'color': ct.DARK_GREEN
            } for track in tracked_objects
        }
        return result

    def update_linear(self, rectangles):
        """
        Update the tracker using linear centroid-based tracking.

        Parameters
        ----------
        rectangles : ndarray
            Array of bounding boxes in the format [startX, startY, endX, endY].

        Returns
        -------
        dict
            Updated tracked objects.

        """
        self.reset_old_objects()

        if len(rectangles) == 0:
            # No new rectangles, increase disappearance counter
            for objectID in list(self.disappeared.keys()):
                self.disappeared[objectID] += 1
                if self.disappeared[objectID] > self.linear_max_age:
                    self.deregister(objectID)
            return self.objects.copy()

        # Compute centroids
        inputCentroids = np.zeros((len(rectangles), 2), dtype=int)
        for i, (startX, startY, endX, endY) in enumerate(rectangles):
            cX = int((startX + endX) / 2.0)
            cY = int((startY + endY) / 2.0)
            inputCentroids[i] = (cX, cY)

        if len(self.objects) == 0:
            # No existing objects, register all input centroids
            for i in range(len(inputCentroids)):
                self.register(inputCentroids[i], rectangles[i])
            return self.objects.copy()

        # Existing objects, attempt to match
        objectIDs = list(self.objects.keys())
        objectCentroids = np.array([obj['centroid'] for obj in self.objects.values()])
        objectRectangles = np.array([obj['rectangle'] for obj in self.objects.values()])

        # Compute distances
        # Size differences
        objects_Hs = objectRectangles[:, 2] - objectRectangles[:, 0]
        inputs_Hs = rectangles[:, 2] - rectangles[:, 0]
        h_D = np.abs(objects_Hs[:, np.newaxis] - inputs_Hs[np.newaxis, :])

        objects_Ws = objectRectangles[:, 3] - objectRectangles[:, 1]
        inputs_Ws = rectangles[:, 3] - rectangles[:, 1]
        w_D = np.abs(objects_Ws[:, np.newaxis] - inputs_Ws[np.newaxis, :])

        hw_D = np.minimum(h_D, w_D)
        centroid_D = dist.cdist(objectCentroids, inputCentroids)

        total_weight = self.hw_dist_weight + self.center_dist_weight
        D = (hw_D * self.hw_dist_weight + centroid_D * self.center_dist_weight) / total_weight

        usedRows = set()
        usedCols = set()
        for row in D.min(axis=1).argsort():
            col = D.argmin(axis=1)[row]
            if row in usedRows or col in usedCols:
                continue
            if D[row, col] > self.linear_max_distance:
                continue
            if centroid_D[row, col] > self.linear_max_distance_center:
                continue
            if hw_D[row, col] > self.linear_max_distance_hw:
                continue

            max_size = max(
                objectRectangles[row][2] - objectRectangles[row][0],
                objectRectangles[row][3] - objectRectangles[row][1],
                rectangles[col][2] - rectangles[col][0],
                rectangles[col][3] - rectangles[col][1]
            )
            if centroid_D[row, col] / max_size > self.linear_max_relative_distance:
                continue

            objectID = objectIDs[row]
            self.objects[objectID]['centroid_history'].append(inputCentroids[col])
            self.objects[objectID]['last_update'] = datetime.now()
            self.objects[objectID]['appearances'] = len(self.objects[objectID]['centroid_history'])
            self.objects[objectID]['centroid'] = inputCentroids[col]
            self.objects[objectID]['rectangle'] = list(rectangles[col][:4])
            self.disappeared[objectID] = 0

            usedRows.add(row)
            usedCols.add(col)

        unusedRows = set(range(D.shape[0])) - usedRows
        unusedCols = set(range(D.shape[1])) - usedCols

        # Increase disappearance counter for unused rows (objects)
        for row in unusedRows:
            objectID = objectIDs[row]
            self.disappeared[objectID] += 1
            if self.disappeared[objectID] > self.linear_max_age:
                self.deregister(objectID)

        # Register new objects for unused columns (input detections)
        for col in unusedCols:
            self.register(inputCentroids[col], rectangles[col])

        # Return copy of objects to avoid side-effects
        return self.objects.copy()

    def get_object_appearances(self, object_id):
        """
        Get the number of appearances of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        int
            Number of appearances.

        """
        if object_id in self.objects:
            return self.objects[object_id]['appearances']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['appearances']
        else:
            return 0

    def get_in_zone_history_deque(self, object_id):
        """
        Get the in-zone history deque of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        deque
            In-zone history deque.

        """
        if object_id in self.objects:
            return self.objects[object_id]['in_zone_history']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['in_zone_history']
        else:
            return deque()

    def get_in_zone_history(self, object_id):
        """
        Get the in-zone history list of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        list
            In-zone history list.

        """
        return list(self.get_in_zone_history_deque(object_id))

    def get_in_zone_total_seconds(self, object_id):
        """
        Get the total seconds an object has been in the zone, excluding any ongoing intervals.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        int
            Total seconds in zone.

        """
        if object_id in self.objects:
            return self.objects[object_id]['in_zone_total_seconds']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['in_zone_total_seconds']
        else:
            return 0

    def set_in_zone_total_seconds(self, object_id, value):
        """
        Set the total seconds an object has been in the zone.

        Parameters
        ----------
        object_id : int
            The ID of the object.
        value : int
            Total seconds to set.

        """
        if object_id in self.objects:
            self.objects[object_id]['in_zone_total_seconds'] = value
        elif object_id in self.objects_history:
            self.objects_history[object_id]['in_zone_total_seconds'] = value

    def get_in_zone_total_seconds_additional(self, object_id):
        """
        Get the total seconds an object has been in the zone, including ongoing intervals.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        int
            Total seconds in zone including ongoing intervals.

        """
        total_seconds = self.get_in_zone_total_seconds(object_id)
        in_zone_history = self.get_in_zone_history(object_id)
        if in_zone_history and len(in_zone_history[-1]) == 1:
            # Last interval is open, include current duration
            additional_seconds = (datetime.now() - in_zone_history[-1][0]).total_seconds()
            total_seconds += int(additional_seconds)
        return total_seconds

    def maybe_close_in_zone_interval(self, object_id):
        """
        Close the in-zone interval if it is still open.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        """
        history_deque = self.get_in_zone_history_deque(object_id)
        if history_deque and len(history_deque[-1]) == 1:
            last_start = history_deque[-1][0]
            now = datetime.now()
            duration = (now - last_start).total_seconds()
            current_total = self.get_in_zone_total_seconds(object_id)
            self.set_in_zone_total_seconds(object_id, current_total + int(duration))
            history_deque[-1].append(now)

    def update_in_zone_history(self, in_zone_objects):
        """
        Update the in-zone history for the provided objects.

        Parameters
        ----------
        in_zone_objects : list of dict
            List of objects currently in the zone.

        """
        current_object_ids = set()
        for obj in in_zone_objects:
            track_id = obj[ct.TRACK_ID]
            current_object_ids.add(track_id)
            history_deque = self.get_in_zone_history_deque(track_id)
            if not history_deque or len(history_deque[-1]) == 2:
                # Start a new interval
                history_deque.append([datetime.now()])

        # For objects not in zone anymore, close the interval if open
        for object_id in self.objects.keys():
            if object_id not in current_object_ids:
                self.maybe_close_in_zone_interval(object_id)

    def get_object_history(self, object_id):
        """
        Get the centroid history of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        list
            List of centroids.

        """
        if object_id in self.objects:
            return list(self.objects[object_id]['centroid_history'])
        elif object_id in self.objects_history:
            return list(self.objects_history[object_id]['centroid_history'])
        else:
            return []

    def get_object_max_movement(self, object_id, steps=None, method='l2'):
        """
        Get the maximum distance the object has moved from its original position.

        Parameters
        ----------
        object_id : int
            The ID of the object.
        steps : int or None, optional
            Number of recent steps to consider. If None, consider all history.
        method : str, optional
            Distance metric to use ('l1' or 'l2').

        Returns
        -------
        float
            Maximum movement distance.

        """
        centroids = self.get_object_history(object_id)
        if not centroids:
            return 0.0

        if steps is not None and isinstance(steps, int):
            centroids = centroids[-steps:]
            original_position = np.array(centroids[0])
        else:
            original_position = np.array(self.get_original_position(object_id))

        centroids_array = np.array(centroids)
        if method == 'l1':
            distances = np.sum(np.abs(centroids_array - original_position), axis=1)
        else:  # Default to 'l2'
            distances = np.linalg.norm(centroids_array - original_position, axis=1)

        return np.max(distances)

    def get_object_type_history(self, object_id):
        """
        Get the type history summary of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        dict
            Type history summary.

        """
        if object_id in self.objects:
            return self.objects[object_id]['type_history']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['type_history']
        else:
            return {'total': 0}

    def get_object_type_history_deque(self, object_id):
        """
        Get the type history deque of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        deque
            Type history deque.

        """
        if object_id in self.objects:
            return self.objects[object_id]['type_history_deque']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['type_history_deque']
        else:
            return deque()

    def add_to_type_history(self, inferences):
        """
        Update the type history with new inferences.

        Parameters
        ----------
        inferences : list of dict
            List of inferences containing object IDs and types.

        """
        for inference in inferences:
            track_id = inference[ct.TRACK_ID]
            obj_type = inference[ct.TYPE]
            type_history_deque = self.get_object_type_history_deque(track_id)
            type_history_deque.append(obj_type)
            type_history = self.get_object_type_history(track_id)
            type_history['total'] += 1
            type_history[obj_type] = type_history.get(obj_type, 0) + 1

    def get_most_seen_type(self, object_id):
        """
        Get the most frequently observed type for an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        str
            The most frequently observed type.

        """
        type_history = self.get_object_type_history(object_id)
        if not type_history or type_history['total'] == 0:
            return ''
        return max(
            (key for key in type_history if key != 'total'),
            key=lambda k: type_history[k],
            default=''
        )

    def get_class_count(self, object_id, class_name, return_complement=False, return_ratio=False):
        """
        Get the count or ratio of times an object was of a specified class.

        Parameters
        ----------
        object_id : int
            The ID of the object.
        class_name : str or list of str
            The class name(s) to check.
        return_complement : bool, optional
            If True, return count/ratio of times the object was not of the specified class.
        return_ratio : bool, optional
            If True, return the ratio instead of the count.

        Returns
        -------
        int or float
            The count or ratio.

        """
        type_history = self.get_object_type_history(object_id)
        total = type_history.get('total', 0)
        if total == 0:
            return 0.0 if return_ratio else 0

        if isinstance(class_name, list):
            count = sum(type_history.get(name, 0) for name in class_name)
        else:
            count = type_history.get(class_name, 0)

        if return_complement:
            count = total - count

        if return_ratio:
            return count / total
        else:
            return count

    def get_non_class_count(self, object_id, class_name):
        """
        Get the count of times an object was not of a specified class.

        Parameters
        ----------
        object_id : int
            The ID of the object.
        class_name : str or list of str
            The class name(s) to check.

        Returns
        -------
        int
            The count.

        """
        return self.get_class_count(object_id, class_name, return_complement=True)

    def get_class_ratio(self, object_id, class_name):
        """
        Get the ratio of times an object was of a specified class.

        Parameters
        ----------
        object_id : int
            The ID of the object.
        class_name : str or list of str
            The class name(s) to check.

        Returns
        -------
        float
            The ratio.

        """
        return self.get_class_count(object_id, class_name, return_ratio=True)

    def get_original_position(self, object_id):
        """
        Get the original centroid position of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        ndarray
            The original centroid position.

        """
        if object_id in self.objects:
            return self.objects[object_id]['original_position']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['original_position']
        else:
            return np.array([0, 0])

    def get_last_rectangle(self, object_id):
        """
        Get the last known rectangle of an object.

        Parameters
        ----------
        object_id : int
            The ID of the object.

        Returns
        -------
        list
            The last rectangle in the format [startX, startY, endX, endY].

        """
        if object_id in self.objects:
            return self.objects[object_id]['rectangle']
        elif object_id in self.objects_history:
            return self.objects_history[object_id]['rectangle']
        else:
            return [0, 0, 0, 0]


if __name__ == "__main__":

  eng = CentroidObjectTracker()




from naeural_core.local_libraries.model_server_v2 import FlaskWorker

_CONFIG = {
  'ADD' : 7
}

class FakeWorker(FlaskWorker):

  """
  Example implementation of a worker

  Obs: as the worker runs on thread, then no prints are allowed; use `_create_notification` and see all the notifications
       when calling /notifications of the server.
  """

  def __init__(self, **kwargs):
    super(FakeWorker, self).__init__(prefix_log='[FAKEW]', **kwargs)
    return


  ### defined properies for worker parameters;
  @property
  def cfg_add(self):
    ### this one is taken directly from default configuration `_CONFIG`
    return self.config_worker['ADD']

  @property
  def cfg_minus(self):
    ### this one can come in upstream configuraion
    return self.config_worker.get('MINUS', None)

  def _load_model(self):
    ### see docstring in parent
    ### abstract method implementation: no model to be loaded
    return

  def _pre_process(self, inputs):
    ### see docstring in parent
    ### abstract method implementation: parses the request inputs and keep the value for 'INPUT_VALUE'
    if 'INPUT_VALUE' not in inputs.keys():
      raise ValueError('INPUT_VALUE should be defined in inputs')

    s = inputs['INPUT_VALUE']
    return s

  def _predict(self, prep_inputs):
    ### see docstring in parent
    ### abstract method implementation: "in-memory model :)" that adds and subtracts.
    self._create_notification(
      notif='log',
      msg='Predicting on usr_input: {}'.format(prep_inputs)
    )

    if not self.cfg_minus:
      res = '{}+{}={} PREDICTED'.format(prep_inputs, self.cfg_add, int(prep_inputs) + self.cfg_add)
    else:
      res = '{}+{}-{}={} PREDICTED'.format(prep_inputs, self.cfg_add, self.cfg_minus, int(prep_inputs) + self.cfg_add - self.cfg_minus)
    return res

  def _post_process(self, pred):
    ### see docstring in parent
    ### abstract method implementation: packs the endpoint answer that will be jsonified
    return {'output_value': pred}



from functools import partial
from collections import OrderedDict
import numpy as np

import tensorflow as tf
import tensorflow.keras.backend as K
import pickle

from tensorflow.keras import initializers, regularizers, constraints

VER = '0.7.5'


class SinCosEncodingLayer(tf.keras.layers.Layer):

  def __init__(self, nr_classes, **kwargs):
    super(SinCosEncodingLayer, self).__init__(**kwargs)
    self.nr_classes = nr_classes
    return

  def call(self, inputs):
    tf_sin_enc = tf.expand_dims(
      tf.keras.backend.sin(2 * np.pi * inputs / self.nr_classes),
      -1
    )

    tf_cos_enc = tf.expand_dims(
      tf.keras.backend.cos(2 * np.pi * inputs / self.nr_classes),
      -1
    )

    tf_x = tf.keras.layers.concatenate([tf_sin_enc, tf_cos_enc])
    return tf_x


  def get_config(self):
    config = {
        'nr_classes' : self.nr_classes,
        }
    base_config = super(SinCosEncodingLayer, self).get_config()
    cfg = {**base_config, **config}
    return cfg


class RepeatElements(tf.keras.layers.Layer):
  def __init__(self, rep, axis, **kwargs):
    super(RepeatElements, self).__init__(**kwargs)
    self.rep = rep
    self.axis = axis
    return

  # def compute_output_shape(self, input_shape):
  #   input_shape = tensor_shape.TensorShape(input_shape).as_list()
  #   return tensor_shape.TensorShape([input_shape[0], self.n, input_shape[1]])

  def call(self, inputs):
    return K.repeat_elements(inputs, self.rep, self.axis)

  def get_config(self):
    config = {
        'rep' : self.rep,
        'axis': self.axis
        }
    base_config = super(RepeatElements, self).get_config()
    cfg = {**base_config, **config}
    return cfg


class SplitLayer(tf.keras.layers.Layer):
  def __init__(self, num_or_size_splits, axis=0, **kwargs):
    super(SplitLayer, self).__init__(**kwargs)
    self.num_or_size_splits = num_or_size_splits
    self.axis = axis
    return

  def call(self, inputs):
    return tf.split(
      inputs,
      num_or_size_splits=self.num_or_size_splits,
      axis=self.axis
    )

  def get_config(self):
    config = {
        'num_or_size_splits' : self.num_or_size_splits,
        'axis': self.axis
        }
    base_config = super(SplitLayer, self).get_config()
    cfg = {**base_config, **config}
    return cfg


class OneHotLayer(tf.keras.layers.Layer):

  def __init__(self, units, **kwargs):
    self.units = units
    super(OneHotLayer, self).__init__(**kwargs)

  def call(self, inputs):
    x = tf.keras.backend.one_hot(inputs, num_classes=self.units)
    return x

  def get_config(self):
    config = {
        'units' : self.units,
        }
    base_config = super(OneHotLayer, self).get_config()
    cfg = {**base_config, **config}
    return cfg
  
  
class SqueezeLayer(tf.keras.layers.Layer):

  def __init__(self, axis, **kwargs):
    self.axis = axis
    super(SqueezeLayer, self).__init__(**kwargs)

  def call(self, inputs):
    x = tf.keras.backend.squeeze(inputs, axis=self.axis)
    return x

  def get_config(self):
    config = {
        'axis' : self.axis,
        }
    base_config = super(SqueezeLayer, self).get_config()
    cfg = {**base_config, **config}
    return cfg
  
  
class SliceAxis1Layer(tf.keras.layers.Layer):
  def __init__(self, pos, **kwargs):
    self.pos = pos
    super(SliceAxis1Layer, self).__init__(**kwargs)
    
    
  def call(self, inputs):
    x = inputs[:, self.pos]
    return x
  
  
  def get_config(self):
    config = {
        'pos': self.pos,
        }
    base_config = super(SliceAxis1Layer, self).get_config()
    cfg = {**base_config, **config}
    return cfg
  

class SliceLayer(tf.keras.layers.Layer):
  """
   this layer is a universal slicer as it employs tf.gather and determines at 
   runtime the dimension of `None` declared tensors
  """
  def __init__(self, pos, axis, keepdims=False, **kwargs):
    self.pos = pos
    self.axis = axis
    self.keepdims = keepdims
    super(SliceLayer, self).__init__(**kwargs)
    
  def compute_output_shape(self, input_shape):
    output_shape = list(input_shape)
    if self.keepdims:
      output_shape[self.axis] = 1
    else:
      del output_shape[self.axis]
    return output_shape
    
  def call(self, inputs):
    pos = self.pos

    if not isinstance(pos, list):
      if pos < 0:
        shape = tf.shape(inputs)[self.axis]
        pos = shape + self.pos

      if self.keepdims:
        pos = [pos]
    #endif

    x = tf.gather(inputs, indices=pos, axis=self.axis)
    return x
  
  
  def get_config(self):
    config = {
        'pos': self.pos,
        'axis' : self.axis,
        'keepdims' : self.keepdims
        }
    base_config = super(SliceLayer, self).get_config()
    cfg = {**base_config, **config}
    return cfg


class SampledSoftmax(tf.keras.layers.Layer):
  def __init__(self, num_classes, num_sampled=100, num_true=1, 
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               **kwargs):
    self.num_sampled = num_sampled
    self.num_classes = num_classes
    self.num_true = num_true
    
    self.kernel_initializer = initializers.get(kernel_initializer)
    self.bias_initializer = initializers.get(bias_initializer)
    
    self.kernel_regularizer = regularizers.get(kernel_regularizer)
    self.bias_regularizer = regularizers.get(bias_regularizer)
    
    self.kernel_constraint = constraints.get(kernel_constraint)
    self.bias_constraint = constraints.get(bias_constraint)
    
    super(SampledSoftmax, self).__init__(**kwargs)

  def build(self, input_shape):
    dense_shape, _ = input_shape
    self.kernel = self.add_weight(name='kernel',
                                  shape=(self.num_classes, dense_shape[1]),
                                  trainable=True,
                                  initializer=self.kernel_initializer,
                                  regularizer=self.kernel_regularizer,
                                  constraint=self.kernel_constraint)
    self.bias = self.add_weight(name='bias',
                                shape=(self.num_classes,),
                                trainable=True,
                                initializer=self.bias_initializer,
                                regularizer=self.bias_regularizer,
                                constraint=self.bias_constraint)

    super(SampledSoftmax, self).build(input_shape)

  def call(self, inputs_and_labels, training=True):
      inputs, labels = inputs_and_labels

      if training:
        return tf.nn.sampled_softmax_loss(
                 weights=self.kernel,
                 biases=self.bias,
                 labels=labels,
                 inputs=inputs,
                 num_sampled=self.num_sampled,
                 num_classes=self.num_classes,
                 num_true=self.num_true)
      else:
        logits = tf.matmul(inputs, tf.transpose(self.kernel))
        logits = tf.nn.bias_add(logits, self.bias)
        labels_one_hot = tf.one_hot(labels, self.num_classes)
        loss = tf.nn.softmax_cross_entropy_with_logits(
            labels=labels_one_hot,
            logits=logits)
        return [loss, logits]
        
  
  def get_config(self):
    config = {
        'num_sampled' : self.num_sampled,
        'num_classes' : self.num_classes,
        'num_true' : self.num_true,
        'kernel_initializer': initializers.serialize(self.kernel_initializer),
        'bias_initializer': initializers.serialize(self.bias_initializer),
        'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
        'bias_regularizer': regularizers.serialize(self.bias_regularizer),
        'kernel_constraint': constraints.serialize(self.kernel_constraint),
        'bias_constraint': constraints.serialize(self.bias_constraint),
        }
    base_config = super(SampledSoftmax, self).get_config()
    cfg = {**base_config, **config}
    return cfg




class InputTensorsPreparator:
  """
  An alternative to prepare_input_tensors function.
  """
  EMBEDDING = 'Embedding'
  ONEHOT = 'OneHotLayer'
  SINCOS = 'SinCosEncodingLayer'

  def __init__(self, log):
    self.logger = log
    self.EmbLayers = {}
    self.InputTensors = {}

    self.prepare_calls = 0
    return
  
  @staticmethod
  def _convert_shape(shape):
    new_shape = []
    for dim in shape:
      if dim <= -1: dim = None
      new_shape.append(dim)
    return tuple(new_shape)

  @staticmethod
  def _get_dtype(input_name, config_embeddings):
    for emb in config_embeddings:
      if input_name == emb['CONNECTED_TO']:
        mode = InputTensorsPreparator._check_mode(emb['MODE'])
        if mode[0] in [InputTensorsPreparator.EMBEDDING, InputTensorsPreparator.ONEHOT]:
          return tf.int32

    return tf.float32


  @staticmethod
  def _check_is_inp_for_embedding(input_name, config_embeddings):
    for emb in config_embeddings:
      if input_name == emb['CONNECTED_TO']:
        return True
    return False

  @staticmethod
  def _get_dict_key_index(dct, key):
    return list(dct).index(key)

  @staticmethod
  def _check_mode(mode):
    assert ',' in mode
    mode = mode.split(',')

    if mode[0] == InputTensorsPreparator.EMBEDDING:
      assert len(mode) == 3
    if mode[0] in [InputTensorsPreparator.ONEHOT, InputTensorsPreparator.SINCOS]:
      assert len(mode) == 2

    return mode
    
  def prepare(self, config_inputs, config_embeddings=None):
    """
    this function returns the model inputs and the post-inputs concatenated
    tensor with all the pre-req for each individual input (one-hots, embeddings etc.)
    
    inputs:
      config_inputs: list that contains definition of graph inputs where each element is a 
                     different tensor (in the order of the actual inputs) represented by a dict
                     with the folllowing possible keys:
                       - "NAME"
                       - "SHAPE" / "BATCH_SHAPE": specified as a list, in which -1 elements
                                                  will be converted to None
                       - "IS_FEEDABLE": [optional] whether the tensor will be fed with data or not
                       
      config_embeddings: list that contains the definition of the embeddings. Each element is a different layer
                         represented by a dict with the following possible keys:
                           - "CONNECTED_TO": the name of the input to which is connected
                           - "NAME" : [optional]. If not specified the name of the emb layer will be
                                      `config_embeddings['CONNECTED_TO'] + '_emb'`
                           - "MODE" :
                              - if embeddings: 'Embedding,<input_dim>,<output_dim>';
                              - if one-hot   : 'OneHotLayer,<nr_classes>';
                              - if sincos    : 'SinCosEncodingLayer,<nr_classes>';

                           - "TRAINABLE": [optional] [used only for embeddings].
                                          Whether the embeddings matrix is trainable or not
                           - "EMB_MATRIX_PATH": [optional] [used only for embeddings].
                                                Path to pre-trained embeddings matrix
                           - "USE_DRIVE": [optional] whether to load the embeddings matrix from the '_data' folder or not
                           - "SQUEEZE_AXIS": [optional] specified if it is necessary to squeeze the tensor
                                             after the encoding
                           - "REUSE": [optional] the name of reused embedding layer. It is used manily when the
                                      architecture is more complex (encoder-decoder, for example)
                           - "BOTTLENECK": [optional] the number of units of the Dense layer that bottlenecks the embedding
    """
    
    all_feats = []
    self.prepare_calls += 1
    if config_embeddings is None:
      config_embeddings = []
    
    ################## INPUTS PROCESSING ##################
    crt_inputs = OrderedDict()
    for _idx, cfg_inp in enumerate(config_inputs):
      name = cfg_inp['NAME']
      is_feedable = cfg_inp.get('IS_FEEDABLE', True)
      dtype = self._get_dtype(name, config_embeddings)
      is_input_for_emb = self._check_is_inp_for_embedding(name, config_embeddings)

      has_batch_shape = 'BATCH_SHAPE' in cfg_inp
      if has_batch_shape:
        tf_x = tf.keras.layers.Input(
          name=name, dtype=dtype,
          batch_shape=self._convert_shape(cfg_inp['BATCH_SHAPE']),
        )
      else:
        tf_x = tf.keras.layers.Input(
          name=name, dtype=dtype,
          shape=self._convert_shape(cfg_inp['SHAPE']),
        )
      #endif
      
      crt_inputs[name] = tf_x
      
      if is_feedable:
        k = len(self.InputTensors) + 1
        self.InputTensors[k] = tf_x
      
      if not is_input_for_emb:
        all_feats.append((_idx,tf_x))
    #endfor
    
    ################## EMBEDDINGS PROCESSING ##################


    for cfg_emb in config_embeddings:
      weights = None
      connected_to = cfg_emb['CONNECTED_TO']
      trainable = cfg_emb.get('TRAINABLE', True)
      squeeze_axis = cfg_emb.get('SQUEEZE_AXIS', None)
      bottleneck_units = cfg_emb.get('BOTTLENECK', None)
      reuse = cfg_emb.get('REUSE', None)
      bool_reuse = not ((reuse is None) or (reuse == ''))

      _idx = self._get_dict_key_index(dct=crt_inputs, key=connected_to)
      mode = self._check_mode(cfg_emb['MODE'])

      name = cfg_emb.get('NAME', None)

      if name is None:
        if mode[0] == InputTensorsPreparator.EMBEDDING:
          name = connected_to + '_emb_{}_{}'.format(mode[1], mode[2])
        elif mode[0] == InputTensorsPreparator.ONEHOT:
          name = connected_to + '_oh_{}'.format(mode[1])
        else: #mode[0] == InputTensorsPreparator.SINCOS
          name = connected_to + '_sincos_{}'.format(mode[1])
      #endif

      lyr_name = name.lower()
  
      if 'EMB_MATRIX_PATH' in cfg_emb and mode[0] == InputTensorsPreparator.EMBEDDING:
        if cfg_emb['EMB_MATRIX_PATH'] != '':
          path = cfg_emb['EMB_MATRIX_PATH']
          use_drive = cfg_emb.get('USE_DRIVE', False)
          if use_drive:
            path = self.logger.get_data_file(path)

          weights = np.load(path)
          assert len(weights.shape) == 2
        #endif
      #endif
      
      if not bool_reuse:
        if mode[0] == InputTensorsPreparator.EMBEDDING:
          if weights is None:
            Emb = tf.keras.layers.Embedding(
              input_dim=int(mode[1]), output_dim=int(mode[2]),
              trainable=trainable, name=lyr_name
            )
          else:
            initializer = tf.keras.initializers.Constant(weights)
            Emb = tf.keras.layers.Embedding(
              input_dim=int(mode[1]), output_dim=int(mode[2]),
              embeddings_initializer=initializer,
              trainable=trainable, name=lyr_name
            )
        elif mode[0] == InputTensorsPreparator.ONEHOT:
          Emb = OneHotLayer(units=int(mode[1]), name=lyr_name)
        else:  # mode[0] == InputTensorsPreparator.SINCOS
          Emb = SinCosEncodingLayer(nr_classes=int(mode[1]), name=lyr_name)
        
        tf_emb = Emb(crt_inputs[connected_to])
        self.EmbLayers[name] = Emb
      else:
        Emb = self.EmbLayers[reuse]
        lyr_name = Emb.name
        tf_emb = Emb(crt_inputs[connected_to])
      #endif

      if squeeze_axis is not None:
        tf_emb = SqueezeLayer(axis=squeeze_axis, name=lyr_name+'_sqz_{}'.format(self.prepare_calls))(tf_emb)
      #endif

      if bottleneck_units is not None:
        tf_emb = tf.keras.layers.Dense(
          units=bottleneck_units,
          name=lyr_name+'_bneck{}'.format(bottleneck_units),
          use_bias=False
        )(tf_emb)
      #endif

      all_feats.append((_idx, tf_emb))
    #endfor


    all_feats = sorted(all_feats, key=lambda x: x[0])
    all_feats = list(zip(*all_feats))
    all_feats = list(all_feats[1])

    return list(crt_inputs.values()), all_feats


def prepare_input_tensors(log, lst_inputs, oh_bneck=None, DEBUG=False):
  """
  this function returns the model inputs and the post-inputs concatenated
  tensor with all the pre-req for each individual input (one-hots, etc)
  
  lst_inputs: list that contains definition of graph inputs where each element is a 
  different tensor (in the order of the actual inputs) represented by a dict where 
  and each key defines the name, tf dtype, number of classes if one-hot is required. 
    For example lets suppose we have three different inputs in the following order: 
        ID that will be embedding encoded (by default if no number of classes
                                           is provided 'EMB_SIZE' tuple is provided), 
        LOCATION that will be one-hot encoded
        VALUE that is a simple float
            lst_inputs = [
                            {
                                'NAME': 'ID',
                                'SHAPE' : [-1] # -1 means None
                                'DTYPE':tf.int32 # this can also be a string such as "int32"
                                'EMB_SIZE': (500, 16),
                            },
                            {
                                'NAME': 'LOCATION',
                                'SHAPE' : [-1, 1] # (None, 1)
                                'DTYPE': tf.int32,
                                'NUM_CLASSES': 30,
                            },
                            {
                                'NAME': 'VALUE',
                                'SHAPE' : [-1, 2] # (None, 2)
                                'DTYPE': tf.float32,
                            } 
                          ]
            
        oh_bneck : (None) if a number it will put a bottleneck dense(units=oh_bneck) after each one-hot encoder
            
    Returns:
      tuple (inputs, post_inputs) the list of input tensors and the post_input ones
      
      
      
  shared-embeddings inputs are handled in InputTensorsPreparator
  """
  if (log is None) or not hasattr(log, '_logger'):
    raise ValueError("Loggger object is invalid: {}".format(log))

  input_list = []
  post_input_tensors = []
  if oh_bneck is not None:
    if type(oh_bneck) is not int or oh_bneck <= 0:
      raise ValueError("One-hot bottleneck param `ohb` must be positive integer. Received {}".format(oh_bneck))
  for dct_inp in lst_inputs:
    inp_type = dct_inp['DTYPE']
    inp = dct_inp['NAME']
    inp_shape = dct_inp['SHAPE']
    inp_tuple = ()
    for pos in inp_shape:
      inp_tuple += (None,) if pos == -1 else (pos,)
    if type(inp_type) is str:
      inp_type = tf.as_dtype(inp_type)
    last_pos_inp = inp_tuple[-1]
    inp_name = 'inp_'+inp
    if DEBUG:
      log.P("  prepare_input_tensors: name:{}  type:{}  shape:{}  ".format(
          inp_name, inp_type, inp_tuple))
    tf_inp = tf.keras.layers.Input(inp_tuple, dtype=inp_type, name=inp_name)
    input_list.append(tf_inp)
    tf_x = tf_inp
    if inp_type == tf.int32:
      if 'NUM_CLASSES' in dct_inp:
        num_cls = dct_inp['NUM_CLASSES']
        #tf.keras.layers.Lambda(function=lambda x: one_hot(x, num_classes=num_cls), name=inp_name+'_oh{}'.format(num_cls))(tf_x)
        tf_x = OneHotLayer(units=num_cls, name=inp_name+'_oh{}'.format(num_cls))(tf_x)
      elif 'EMB_SIZE' in dct_inp:
        vocab_size, emb_size = dct_inp['EMB_SIZE']
        if 'TRAINABLE' not in dct_inp:
          dct_inp['TRAINABLE'] = True
        if 'WEIGHTS' in dct_inp:
          with open(dct_inp['WEIGHTS'], 'rb') as f:
            weights = pickle.load(f)
            initializer = tf.keras.initializers.Constant(weights)
          l_emb = tf.keras.layers.Embedding(vocab_size, emb_size, embeddings_initializer=initializer, trainable=dct_inp['TRAINABLE'],
              name=inp_name+'_emb_{}_{}'.format(vocab_size, emb_size))
        else:
          l_emb = tf.keras.layers.Embedding(vocab_size, emb_size, trainable=dct_inp['TRAINABLE'],
              name=inp_name+'_emb_{}_{}'.format(vocab_size, emb_size))
        tf_x = l_emb(tf_x)
      else:
        raise ValueError('INT32 input received without classes/embeds definition')  
      if last_pos_inp == 1:
        # we should squeeze the singleton dim
        # tf.keras.layers.Lambda(function=lambda x: squeeze(x, axis=-2), name=inp_name+'_sqz')(tf_x)
        tf_x = SqueezeLayer(axis=-2, name=inp_name+'_sqz')(tf_x)
        if oh_bneck is not None:
          tf_x = tf.keras.layers.Dense(units=oh_bneck, 
                                       name=inp_name+'_oh_bneck{}'.format(oh_bneck),
                                       use_bias=False)(tf_x)
        
    post_input_tensors.append(tf_x)
        
  return input_list, post_input_tensors


def _ts_causal_conv_res_module_he(tf_x, kernel, filters, dilation, bn, activ, drop, bname, nr_reps=2):
  tf_skip = tf_x
  inp_filters = tf_x.shape.as_list()[-1]
  if filters != inp_filters:
    tf_skip = tf.keras.layers.Conv1D(filters, 
                                     kernel_size=1, 
                                     padding='causal',
                                     name=bname+'_ident')(tf_skip)
    tf_x = tf_skip

  recv_field = 0
  bname += '_d'+str(dilation)
  for L in range(1, nr_reps+1):
    if bn:
      tf_x = tf.keras.layers.BatchNormalization(name=bname+'_bnpre{}'.format(L))(tf_x)
    tf_x = tf.keras.layers.Activation(activ, name=bname+'_{}{}'.format(activ,L))(tf_x)
    if drop > 0:
      tf_x = tf.keras.layers.Dropout(rate=drop, name=bname+'_drp{}_{:.1f}'.format(L,drop))(tf_x)
    tf_x = tf.keras.layers.Conv1D(filters=filters,
                                  kernel_size=kernel,
                                  padding='causal',
                                  dilation_rate=dilation,
                                  name=bname+'_cnv{}'.format(L))(tf_x)
    recv_field += (kernel - 1) * dilation
  tf_out = tf.keras.layers.add([tf_skip, tf_x], name=bname+'_res_rf_{}'.format(recv_field))
  return tf_out, recv_field


def _ts_causal_conv_res_module_res(tf_x, kernel, filters, dilation, bn, activ, drop, bname, nr_reps=2):

  tf_skip = tf_x
  inp_filters = tf_x.shape.as_list()[-1]
  recv_field = 0
  if filters != inp_filters:
    tf_skip = tf.keras.layers.Conv1D(filters, 
                                     kernel_size=1, 
                                     padding='causal',
                                     name=bname+'_ident')(tf_skip)
  
  bname += '_d'+str(dilation)

  for L in range(1, nr_reps):
    tf_x = tf.keras.layers.Conv1D(filters=filters,
                                  kernel_size=kernel,
                                  padding='causal',
                                  dilation_rate=dilation,
                                  name=bname+'_cnv{}'.format(L))(tf_x)
    recv_field += (kernel - 1) * dilation
    if bn:
      tf_x = tf.keras.layers.BatchNormalization(name=bname+'_bnpre{}'.format(L))(tf_x)
    tf_x = tf.keras.layers.Activation(activ, name=bname+'_{}{}'.format(activ,L))(tf_x)
  
  tf_x = tf.keras.layers.Conv1D(filters=filters,
                                kernel_size=kernel,
                                padding='causal',
                                dilation_rate=dilation,
                                name=bname+'_cnv{}'.format(nr_reps))(tf_x)
  recv_field += (kernel - 1) * dilation
  if bn:
    tf_x = tf.keras.layers.BatchNormalization(name=bname+'_bnpre{}'.format(nr_reps))(tf_x)
    
  tf_x = tf.keras.layers.add([tf_skip, tf_x], name=bname+'_res')
  
  tf_x = tf.keras.layers.Activation(activ, name=bname+'_{}{}_rf_{}'.format(activ,'out',recv_field))(tf_x)
  
  if drop > 0:
    tf_x = tf.keras.layers.Dropout(rate=drop, name=bname+'_drp_{:.1f}'.format(drop))(tf_x)
  tf_out = tf_x  
  return tf_out, recv_field


def _ts_causal_conv_res_module_bai(tf_x, kernel, filters, dilation, bn, activ, drop, bname, nr_reps=2):
  tf_skip = tf_x
  inp_filters = tf_x.shape.as_list()[-1]
  recv_field = 0
  if filters != inp_filters:
    tf_skip = tf.keras.layers.Conv1D(filters, 
                                     kernel_size=1, 
                                     padding='causal',
                                     name=bname+'_ident')(tf_skip)
  
  bname += '_d'+str(dilation)
  for L in range(1, nr_reps+1):
    tf_x = tf.keras.layers.Conv1D(filters=filters,
                                  kernel_size=kernel,
                                  padding='causal',
                                  dilation_rate=dilation,
                                  name=bname+'_cnv{}'.format(L))(tf_x)
    recv_field += (kernel - 1) * dilation
    if bn:
      tf_x = tf.keras.layers.BatchNormalization(name=bname+'_bnpre{}'.format(L))(tf_x)
    tf_x = tf.keras.layers.Activation(activ, name=bname+'_{}{}'.format(activ,L))(tf_x)
    if drop > 0:
      tf_x = tf.keras.layers.Dropout(rate=drop, name=bname+'_drp{}_{:.1f}'.format(L,drop))(tf_x)
  
    
  tf_x = tf.keras.layers.add([tf_skip, tf_x], name=bname+'_res')
  
  tf_x = tf.keras.layers.Activation(activ, name=bname+'_{}{}_rf'.format(activ,'out',recv_field))(tf_x)
  
  if drop > 0:
    tf_x = tf.keras.layers.Dropout(rate=drop, name=bname+'_drp_{}'.format(drop))(tf_x)

  tf_out = tf_x  
  return tf_out, recv_field


def _ts_causal_conv_res_module(tf_x, 
                                        kernel, 
                                        filters, 
                                        dilation, 
                                        bn, 
                                        activ, 
                                        drop, 
                                        bname, 
                                        nr_reps,
                                        res,
                                        default_dilation=0,
                                        progressive_dilation=False,
                                        ):
  if default_dilation != 0:
    dilation = default_dilation
  
  if progressive_dilation:
    _pow = int(np.log2(dilation))
  recv_field = 0
  if res:
    tf_skip = tf_x
    inp_filters = tf_x.shape.as_list()[-1]
    if filters != inp_filters:
      tf_skip = tf.keras.layers.Conv1D(filters, 
                                       kernel_size=1, 
                                       padding='causal',
                                       name=bname+'_ident')(tf_skip)
  
  for L in range(1, nr_reps+1):
    if progressive_dilation:
      dilation = 2 ** _pow
      _pow += 1
    recv_field += (kernel - 1) * dilation
    tf_x = tf.keras.layers.Conv1D(filters=filters,
                                  kernel_size=kernel,
                                  padding='causal',
                                  dilation_rate=dilation,
                                  name=bname+'_cnv{}_d{}_rf{}'.format(L,dilation,recv_field))(tf_x)
    

    if bn:
      tf_x = tf.keras.layers.BatchNormalization(name=bname+'_bnpre{}'.format(L))(tf_x)
      
    if activ not in [None, '', 'none', 'linear']:
      tf_x = tf.keras.layers.Activation(activ, name=bname+'_{}{}'.format(activ,L))(tf_x)

    if drop > 0:
      tf_x = tf.keras.layers.Dropout(rate=drop, name=bname+'_drp{}_{:.1f}'.format(L,drop))(tf_x)
  
    
  if res:
    tf_x = tf.keras.layers.add([tf_skip, tf_x], name=bname+'_res')
  
  tf_out = tf_x  
  return tf_out, recv_field



dct_FUNCS = {
    'he' : _ts_causal_conv_res_module_he,
    'ba' : _ts_causal_conv_res_module_bai,
    'rs' : _ts_causal_conv_res_module_res,
    'Lf' : partial(_ts_causal_conv_res_module, res=True),
    'Lr' : partial(_ts_causal_conv_res_module, res=True, default_dilation=1),
    'Ld' : partial(_ts_causal_conv_res_module, res=False),
    'Lc' : partial(_ts_causal_conv_res_module, res=False, default_dilation=1),
    'Lp' : partial(_ts_causal_conv_res_module, res=True, progressive_dilation=1),
    }
  
def _ts_causal_conv_res_module(tf_x, kernel, filters, dilation, bn, activ, drop, bname, 
                               architecture='ba',
                               nr_reps=2):
  func = dct_FUNCS[architecture]
  tf_out, recv_field = func(tf_x, kernel, filters, dilation, bn, activ, drop, bname, nr_reps)
  return tf_out, recv_field
  


def timeseries_causal_conv_module(
                        log,
                        tf_input,
                        kernels=[  31,  7,    7,   365], 
                        filters=[ 512, 256,  512,  256,],
                        factors=[ 1.0, 1.0,  1.0,  1.0,],
                        depths= [   1,   6,   1,    1],
                        drops=  [   0,   0,   0,    0],
                        repeats=[   1,   2,   1,    1], # do not repeat many times!
                        bns=    [   0,   0,   0,    0], 
                        layouts=['Lc','Lf',  'Lc', 'Lc'],
                        activs=['selu', 'selu', 'selu', 'selu'], 
                        return_model=False,
                        skip=True,
                        name_suffix='',
                        DEBUG=False,
                        verbose=2,
                ):
  """
   Creates a causal convolution inception module with `len(kernels)` columns 
   for time-series modeling
   
   OBSERVATION: please run libraries/nn/utils.py `__main__` code to get some 
   intuition on the functionalities of this method
   
   inputs:
     
     `log`  :  Logger object
     
     `tf_input` : input tensor for the module
     
     `filters`: list of filters for each kernel default

     `factors`: the nr filters on each column should be increased with the depth
                (i.e. each residual block should have more filters than the previous one).
                This param specifies for each column the scale factor for the filters param.

                Could be:
                  * list - a factor for each column
                  * float - same factor is applied on all columns

     
     `kernels`: kernel size (temporal window) 
     
     `depths` : how many residual blocks per column
     
             `kernels=[  2,  3,  4,  4,  7,],`  31  / 61 / 91 / 373  days receptive fields
             `filters=[128,128,128,128,128,],` 
             `depths= [  4,  4,  4,  5,  5,],`
             `drops=  [0.2,0.2,0.2,0.2,0.2,],`
             `repeats=[  2,  2,  2,  2,  2,],`
     
     `repeats` : how many convolutions per residual block
     
     `layouts` : architecture definition of each column!
     
     `drops`  : dropout post activation for each column
     
     `bns` :  BatchNorms - True or False for each column     
           
     `activs` : activation defaults to `selu` for each column
     
     `skip` : skip from input

     `verbose` : int, optional
        Logging verbosity. Set 0 for no prints.
        Default is 1.

  returns : 
    
      concatenated tf tensor or the whole model
      
  comments:
    # k/d/r:rcv
    # 2/4/2:31, 2/5/2:63,
    # 3/3/2:29, 3/4/2:61, 
    # 4/4/2:91, 4/5/2:187,
    # 7/2/2:37, 7/3/2:85, 7/4/2:181, 7/5/2:373    
  """
  if DEBUG:
    if tf_input.name[:5].lower() != 'input':
      raise ValueError("In debug mode input tensor must me an actual 'Input' tensor. Received {}".format(
          tf_input.name))
  if type(layouts) not in [list, tuple]:
    raise ValueError("Param architecture [list] must define architecture for each column!")
  if type(activs) not in [list, tuple]:
    raise ValueError("Param activs [list] must define activation for each column!")

  if isinstance(factors, float):
    factors = [factors]
  elif isinstance(factors, (list, tuple)):
    pass
  else:
    raise ValueError("Param factors not defined correctly. Please provide `list` or `float`")


  _AVAIL_ARCHITECTURES = list(dct_FUNCS.keys())
  if (log is None) or not hasattr(log, '_logger'):
    raise ValueError("Loggger object is invalid: {}".format(log))
  if len(filters) < len(kernels):
    filters += [filters[-1]] * (len(kernels) - len(filters))
  if len(factors) < len(kernels):
    factors += [factors[-1]] * (len(kernels) - len(factors))
  if len(depths) < len(kernels):
    depths += [depths[-1]] * (len(kernels) - len(depths))
  if len(drops) < len(kernels):
    drops += [drops[-1]] * (len(kernels) - len(drops))
  if len(repeats) < len(kernels):
    repeats += [repeats[-1]] * (len(kernels) - len(repeats))
  if len(bns) < len(kernels):
    bns += [bns[-1]] * (len(kernels) - len(bns))
  if len(layouts) < len(kernels):
    layouts += [layouts[-1]] * (len(kernels) - len(layouts))
  if len(activs) < len(kernels):
    activs += [activs[-1]] * (len(kernels) - len(activs))

  suf = name_suffix
  outputs = []
  columns = []
  summary = []
  lst_receptive_fields = []
  s_recept=''

  if verbose >= 2:
    log.P("  Creating {}{} column Time Convolution Inception Module...".format(
        len(kernels),
        "+1 (skip)" if skip else " (no skip)"), color='y')
  #endif

  lst_par = []
  for i in range(len(kernels)):
    layout = layouts[i]
    s_basename = 'TC_{}'.format(layout)
    if layout not in _AVAIL_ARCHITECTURES:
      raise ValueError("Unknown residual module layout '{}'. Available designs are: {}".format(
          layout, _AVAIL_ARCHITECTURES))
    #endif

    f = filters[i]
    fact = factors[i]
    k = kernels[i]
    drop = drops[i]
    depth = depths[i]
    bn = bns[i]
    nr_reps = repeats[i]
    activ = activs[i]
    tf_x = tf_input
    receptive_field = 1
    s_col_name = '{}{}_k{}_f{}_Col{}'.format(suf, s_basename, k, f, i+1)
    columns.append(s_col_name)
    for j in range(depth):
      dilation = (2 ** j)
      bname = '{}_B{}'.format(s_col_name,j+1)
      #if nr_reps < 2 and architecture.lower() in ['res']:
      #  log.P("  Causal block '{}' must have at least 2 repeats. Increasing to 2.".format(nr_reps))
      #  nr_reps = 2
      tf_x, recv_field = _ts_causal_conv_res_module(
        tf_x=tf_x,
        kernel=k,
        filters=int(f * np.power(fact, j)),
        dilation=dilation,
        bn=bn,
        activ=activ,
        drop=drop,
        bname=bname,
        architecture=layout,
        nr_reps=nr_reps
      )
      receptive_field += recv_field
    #endfor - j in range(depth)

    lst_receptive_fields.append(receptive_field)
    outputs.append(tf_x)    
    n_par = 0

    if DEBUG:
      _m = tf.keras.models.Model(tf_input, tf_x, name='{}block_test'.format(suf))
      n_par = _m.count_params()
      lst_par.append(n_par)
    #endif

    summary.append("    Column '{}' recv field: {} @ {:,} params".format(s_col_name, receptive_field, n_par))
    s_recept += str(receptive_field)+'_'
  #endfor - i in range(len(kernels))
  
  if skip:
    outputs.append(tf_input)
  
  if len(outputs) > 1:
    tf_output = tf.keras.layers.concatenate(outputs, name='{}blk_out_rcf_{}'.format(
                                                  suf, 'skp_' if skip else '')+s_recept)
  else:
    tf_output = tf_x

  if verbose >= 2:
    for _msg in summary:
      log.P(_msg, color='y')

  if DEBUG:
    log.P("    Total {:,} parameters".format(sum(lst_par)), color='y')

  if return_model:
    s_model = 'Model_{}_{}_{}'.format(s_basename, len(kernels), s_recept[:-1])
    m = tf.keras.models.Model(inputs=tf_input, outputs=tf_output,
                              name=s_model)
    return m
  else:
    return tf_output, lst_receptive_fields, outputs
  
  

##### PLEASE PUT MAIN IN DIFFERENT SCRIPT AND IMPORT UTIL METHODS FROM THIS MODULE!
# if __name__ == '__main__':
#   import sys
#   sys.path.append("../")
#  # from naeural_core import Logger
#  # l = Logger(lib_name='LYRUTLS', config_file='libraries/config.txt')
#  # l.clear_all_results()
#   TEST_INPUTS = False
#   TEST_TC1 = False
#   TEST_TC2 = False
#   TEST_SLICE = False
#
#   import cloudpickle
#   with open('oh.dat', 'wb') as f:
#     cloudpickle.dump(OneHotLayer, f)
#
#   if TEST_SLICE:
#     a = np.arange(200).reshape(2,25,4)
#     pos = 0
#     tf_inp = tf.keras.layers.Input((None,4))
#     tf_x = SliceLayer(pos=pos,axis=1)(tf_inp)
#     m = tf.keras.models.Model(tf_inp, tf_x)
#     y = m.predict(a)
#     l.P("output:\n{}".format(y))
#     l.P("truth:\n{}".format(a[:,pos]))
#
#   if TEST_INPUTS:
#     import json
#     with open('inputs_json.txt', 'r') as f:
#       dict_inputs = json.load(f)
#     lst_inputs = dict_inputs['MODEL_INPUTS']
#     if True:
#       inputs, post_inputs = prepare_input_tensors(lst_inputs, DEBUG=True)
#     else:
#       inputs = [tf.keras.layers.Input((None,1), dtype=tf.as_dtype("int32")) for x in range(4)]
#       post_inputs1 = [tf.keras.layers.Embedding(100,16)(x) for x in inputs[:2]]
#       post_inputs2 = [tf.keras.layers.Lambda(lambda x: K.one_hot(x, num_classes=10))(x)
#                       for x in inputs[2:]]
#       post_inputs = post_inputs1 + post_inputs2
#       post_inputs = [tf.keras.layers.Lambda(lambda x: K.squeeze(x, axis=-2))(x)
#                       for x in post_inputs]
#
#     if len(post_inputs) > 1:
#       tf_x = tf.keras.layers.concatenate(post_inputs, name='concat_inputs')
#     else:
#       tf_x = post_inputs[0]
#     tf_out = tf.keras.layers.Dense(1)(tf_x)
#     model = tf.keras.models.Model(inputs=inputs, outputs=tf_out)
#     model.summary()
#     model.save("test_2nd")
#
#   if TEST_TC1:
#
#     tf_inp = tf.keras.layers.Input((10,1))
#     lyr_c1 = tf.keras.layers.Conv1D(filters=1,
#                                    kernel_size=3,
#                                    padding='causal',
#                                    use_bias=False)
#     lyr_c2 = tf.keras.layers.Conv1D(filters=1,
#                                    kernel_size=3,
#                                    padding='causal',
#                                    dilation_rate=2,
#                                    use_bias=False)
#     layers = [lyr_c1, lyr_c2]
#     tf_x1 = lyr_c1(tf_inp)
#     tf_x2 = lyr_c2(lyr_c1(tf_inp))
#     model1 = tf.keras.models.Model(tf_inp, tf_x1)
#     model2 = tf.keras.models.Model(tf_inp, tf_x2)
#     for layer in layers:
#       _w = layer.get_weights()[0]
#       layer.set_weights([np.ones(_w.shape)])
#     x = np.arange(10).reshape(1,10,1)
#     x[:,::3,:] = 0
#     y1 = model1.predict(x)
#     y2 = model2.predict(x)
#     l.P("Demonstrating temporal convolution with `ones` kernels: {}".format(
#       layer.get_weights()[0].tolist()))
#     l.P("Input:\n{}".format(x.ravel()))
#     l.P("First step: output of simple model (k=3):\n{}".format(y1.ravel()))
#     l.P("Second step: output of two lyrs (prev + second layer dilation=2):\n{}".format(y2.ravel()))
#
#   if TEST_TC2:
#
#     # 2/4/2:31, 2/5:63,
#     # 3/3/2:29, 3/4/2:61,
#     # 4/4/2:91, 4/5/2:187,
#     # 7/2/2:37, 7/3/1: 43,  7/3/2:85, 7/4/2:181, 7/5/2: 373
#     # 32/2/1:94
#     # 64/2/1: 190
#     # 128/2/1: 382
#     # 256/2/1: 766
#
#     GOOD = [
#               {
#                   "NAME" : "1x6_full_946",
#                   "PARAMS" :
#                         { # recv field: 946 @ 5,288,192 params
#                             "kernels" : [ 16,],
#                             "filters" : [256,],
#                             "depths"  : [  6,],
#                             "drops"   : [0.2,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lf'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#               {
#                   "NAME" : "1x2_full_382",
#                   "PARAMS" :
#                         { # recv field: 382 @ 8,719,616 params
#                             "kernels" : [128,],
#                             "filters" : [256,],
#                             "depths"  : [  2,],
#                             "drops"   : [0.2,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lf'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#
#
#               {
#                   "NAME" : "1x2_res_nd_31",
#                   "PARAMS" :
#                         { # recv field: 31 @ 1,092,864 params
#                             "kernels" : [ 16,],
#                             "filters" : [256,],
#                             "depths"  : [  2,],
#                             "drops"   : [0.2,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lr'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#
#
#               {
#                   "NAME" : "1x1_cc_nd_31",
#                   "PARAMS" :
#                         { # selu 31 step selu conv
#                             "kernels" : [ 31,],
#                             "filters" : [256,],
#                             "depths"  : [  1,],
#                             "drops"   : [  0,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lc'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#
#               {
#                   "NAME" : "1x1_cc_nd_14",
#                   "PARAMS" :
#                         { #  14 step selu conv
#                             "kernels" : [ 14,],
#                             "filters" : [256,],
#                             "depths"  : [  1,],
#                             "drops"   : [  0,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lc'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#
#               {
#                   "NAME" : "1x1_cc_nd_365",
#                   "PARAMS" :
#                         { #  365 step selu conv
#                             "kernels" : [ 365,],
#                             "filters" : [256,],
#                             "depths"  : [  1,],
#                             "drops"   : [  0,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lc'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#               {
#                   "NAME" : "1x1_cc_nd_732",
#                   "PARAMS" :
#                         { #  732 step linear conv
#                             "kernels" : [ 732,],
#                             "filters" : [256,],
#                             "depths"  : [  1,],
#                             "drops"   : [  0,],
#                             "repeats" : [  1,],
#                             "layouts" : ['Lc'],
#                             "bns"     : [   0],
#                             "activs"  : ['selu'],
#                         },
#               },
#
#               {
#                   "NAME"   : "4xX_cc_nd_732_good",
#                   "PARAMS" :
#                     {
#                             #
#                             "kernels" : [  14,  31,  365, 732,],
#                             "filters" : [ 256, 256, 256, 256,],
#                             "depths"  : [   1,   1,   1,   1,],
#                             "drops"   : [   0,   0,   0,   0,],
#                             "repeats" : [   1,   1,   1,   1,],
#                             "layouts" : ['Lc','Lc','Lc','Lc',],
#                             "bns"     : [   0,   0,   0,   0,],
#                             "activs"  : ['selu','selu','selu','selu']
#                     }
#               },
#         ]
#
#     conv_grid = [
#
#               {
#                   "NAME" : "2X732",
#                   "PARAMS" :
#                         { #  732 step linear conv
#                             "kernels" : [ 31,   7],
#                             "filters" : [512, 256],
#                             "depths"  : [  1,   6],
#                             "drops"   : [  0,   0],
#                             "repeats" : [  1,   2],
#                             "layouts" : ['Lc','Lf'],
#                             "bns"     : [   0,  0],
#                             "activs"  : ['selu','selu'],
#                         },
#               },
#
#               {
#                   "NAME"   : "4cNd732",
#                   "PARAMS" :
#                     {
#                             #
#                             "kernels" : [  14,  31,  365, 732,],
#                             "filters" : [ 256, 256, 256, 256,],
#                             "depths"  : [   1,   1,   1,   1,],
#                             "drops"   : [   0,   0,   0,   0,],
#                             "repeats" : [   1,   1,   1,   1,],
#                             "layouts" : ['Lc','Lc','Lc','Lc',],
#                             "bns"     : [   0,   0,   0,   0,],
#                             "activs"  : ['selu','selu','selu','selu']
#                     }
#               },
#
#
#         ]
#
#     get_model = False
#     bn=True
#     l.P("")
#     l.P("")
#     for i, dparams in enumerate(conv_grid):
#       s_model = "Model__{}".format(dparams['NAME'])
#       l.P(s_model + ':')
#       tf_inp = tf.keras.layers.Input((None,100),)
#       params = dparams['PARAMS']
#       if get_model:
#         m1 = timeseries_causal_conv_module(log=l, tf_input=tf_inp,
#                                            return_model=True,
#                                            DEBUG=True,
#                                            **params
#                                            )
#         tf_x = m1(tf_inp)
#       else:
#         tf_x = timeseries_causal_conv_module(log=l,
#                                              tf_input=tf_inp,
#                                              DEBUG=True,
#                                              **params
#                                              )
#       m = tf.keras.models.Model(tf_inp, tf_x, name=s_model)
#       l.plot_keras_model(m, verbose=False)
#     l.P("Standard :")
#     tf_x = timeseries_causal_conv_module(log=l, tf_input=tf_inp, DEBUG=True)
#     m = tf.keras.models.Model(tf_inp, tf_x, name='std')
#     l.plot_keras_model(m, verbose=False)
#


import numpy as np
from time import time


def intersect(np_box1, np_box2):
  assert np_box1.shape[0] == np_box2.shape[0]
  np_top     = np_box1[:,0] <= np_box2[:,0]
  np_left    = np_box1[:,1] <= np_box2[:,1]
  np_bottom  = np_box1[:,2] >= np_box2[:,2]
  np_right   = np_box1[:,3] >= np_box2[:,3]
  result = np.all([np_top, np_left, np_bottom, np_right], axis=0)
  return result


def get_boxes_overlapp_relations(np_boxes1, np_boxes2):
  """
  This method returns the overlapp mapping between two list of boxes.

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes. (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes. (ex: faces)

  Returns
  -------
  rel_boxes1 : dictionary
    Each box1 item will be maped to a list of box2 items that do intersect with the box1 item
    -> keys will be formed by boxes1 ids
    -> values will be formed by lists of ids from boxes2 that are intersected with the current box2
  rel_boxes2 : dictionary
    Each box2 item will be maped to a list of box1 items that do intersect with the box2 item
    -> keys will be formed by boxes2 ids
    -> values will be formed by lists of ids from boxes1 that are intersected with the current box2

  """
  
  nr_boxes1 = np_boxes1.shape[0]
  nr_boxes2 = np_boxes2.shape[0]
  
  np_boxes1 = np.repeat(np_boxes1, repeats=nr_boxes2, axis=0) #repeat each box1 item to compare it with each box2 item. shape = (nr_boxes1*nr_boxes2, N)
  np_boxes2 = np.vstack([np_boxes2] * nr_boxes1) #for each item in box1, create a replica of box2 items. shape = (nr_boxes2*nr_boxes1, N)
  
  np_overlapps = intersect(np_boxes1, np_boxes2)  #check overlapping boxes. shape = (nr_boxes1*nr_boxes2,)
  
  #create overlapp relations between box1 items and box2 items
  np_overlapps1 = np_overlapps.reshape(nr_boxes1, nr_boxes2) #reshape elements such that each box1 item contains box2 items. shape = (nr_boxes1, nr_boxes2)
  rel_boxes1 = {i: np.argwhere(x).ravel().tolist() for i,x in enumerate(np_overlapps1)} #key represented by box1 item id, value represented by list of box2 ids that overlapp over current box1 item
  
  #create overlapp relations between box1 items and box2 items
  np_overlapps2 = np_overlapps1.T ##reshape such that each box2 item gets its corresponding box1 items. shape = (nr_boxes2, nr_boxes1)
  rel_boxes2 = {i: np.argwhere(x).ravel().tolist() for i,x in enumerate(np_overlapps2)} #key represented by box2 item id, value represented by list of box1 ids that overlapp over current box2 item
  return rel_boxes1, rel_boxes2


def get_non_overlappping_boxes(np_boxes1, np_boxes2):
  """
  This method analyzes two ndarray of TLBR boxes and return only the boxes that are non-overlapping

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes (ex: faces)

  Returns
  -------
  np_box1 : np.ndarray
    Only the initial box1 items that are non overlapping with box2 items
  np_box2 : np.ndarray
    Only the initial box2 items that are non overlapping with box1 items

  """
  
  rel_box1, rel_box2 = get_boxes_overlapp_relations(np_boxes1, np_boxes2)
  
  #extract boxes1 that are not overlapping
  sel_box1 = [k for k,v in rel_boxes1.items() if not v]
  np_box1 = np_boxes1[sel_box1]
  
  #extract boxes2 that are not overlapping
  sel_box2 = [k for k,v in rel_boxes2.items() if not v]
  np_box2 = np_boxes2[sel_box2]
    
  return np_box1, np_box2


def get_only_overlapping_boxes(np_boxes1, np_boxes2):
  """
  This method analyzes two arrays of TLBR boxes and returns only boxes that are intersecting.

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes (ex: faces)

  Returns
  -------
  np_box1 : np.ndarray
    Only the initial box1 items that are overlapping with box2 items
  np_box2 : np.ndarray
    Only the initial box2 items that are overlapping with box1 items

  """
  rel_box1, rel_box2 = get_boxes_overlapp_relations(np_boxes1, np_boxes2)
  
  #extract boxes1 that are overlapping
  sel_box1 = [k for k,v in rel_boxes1.items() if v]
  np_box1 = np_boxes1[sel_box1]
  
  #extract boxes2 that are not overlapping
  sel_box2 = [k for k,v in rel_boxes2.items() if v]
  np_box2 = np_boxes2[sel_box2]
    
  return np_box1, np_box2


def unify_overlapping_boxes(np_boxes1, np_boxes2, box1_priority_on_overlap=True):
  """
  This method analyzes two arrays of TLBR boxes and returns the "union" of those two arrays, based on choosen box priority.
  For example: if `np_boxes1` are persons, `np_boxes2` are faces and `box1_priority_on_overlap` is True, then the method will:
    - keep & return all persons that are non intersecting with a face box
    - keep & return all faces that are non intersecting with a person box
    - keep person in case that person is intersecting a face box
  

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes (ex: faces)
  box1_priority_on_overlap : TYPE, optional
    Choose he box that will be kept in case of intersection

  Returns
  -------
  np_box1 : np.ndarray
    Will contain all non-intersecting box1 items. In case `box1_priority_on_overlap` is set to True, will also contain box1 items that do intersect with box2 items
  np_box2 : np.ndarray
    Will contain all non-intersecting box2 items. In case `box1_priority_on_overlap` is set to False, will also contain box2 items that do intersect with box1 items

  """
  rel_box1, rel_box2 = get_boxes_overlapp_relations(np_boxes1, np_boxes2)
  
  #extract boxes1 that are not overlapping
  sel_box1 = [k for k,v in rel_boxes1.items() if not v]
  
  #extract boxes2 that are not overlapping
  sel_box2 = [k for k,v in rel_boxes2.items() if not v]
  
  if box1_priority_on_overlap:
    sel_box1+= [k for k,v in rel_boxes1.items() if v]
    
  else:
    sel_box2+= [k for k,v in rel_boxes2.items() if v]
  
  np_box1 = np_boxes1[sel_box1]
  np_box2 = np_boxes2[sel_box2]
  
  return np_box1, np_box2


if __name__ == '__main__':
  BOXES1 = np.array([[i*10, 0, (i+1) * 10, 10] for i in range(5)])
  BOXES2 = np.array([[i*10, 20, (i+1) * 10, 30] for i in range(5)] + [[i*5, 0, (i+1) * 5, 5] for i in range(5)])
  
  start = time()
  rel_boxes1, rel_boxes2 = get_boxes_overlapp_relations(BOXES1, BOXES2)
  stop = time()
  print('First: {}'.format(stop - start))

  start = time()
  np_boxes1, np_boxes2 = unify_overlapping_boxes(BOXES1, BOXES2)
  stop = time()
  print('Second: {}'.format(stop - start))


  





#global dependencies
import abc

#local dependencies
from naeural_core.serving.base.base_serving_process import ModelServingProcess as BaseServingProcess
from naeural_core import constants as ct

_CONFIG = {
  **BaseServingProcess.CONFIG,
  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

class BasicTFFunctionServer(BaseServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self.default_input_shape = None
    super(BasicTFFunctionServer, self).__init__(**kwargs)
    return      
  
  @property
  def cfg_gpu_min_memory_mb(self):
    return self.config_model.get(ct.GPU_MIN_MEMORY_MB)
  
  @property
  def cfg_url(self):
    return self.config_model.get(ct.URL, None)
  
  @property
  def cfg_graph(self):
    return self.config_model[ct.GRAPH]
  
  def limit_gpu_memory(self, memory_size_mb):
    # assume running on device 0
    import tensorflow as tf
    gpu = tf.config.experimental.list_physical_devices('GPU')[0]
    cfgs = [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=memory_size_mb)]
    tf.config.experimental.set_virtual_device_configuration(gpu, cfgs)
    return

  ###
  ### BELOW MANDATORY (or just overwritten) FUNCTIONS:
  ###

  
  
  def _startup(self):
    msg = self._setup_model()
    return msg  

  @abc.abstractmethod
  def _setup_model(self):
    raise NotImplementedError()


from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend
import base64
import hashlib
import json

# Generate keys
private_key = ec.generate_private_key(ec.SECP256K1(), default_backend())
public_key = private_key.public_key()

# Serialize public key in X9.62 compressed point format
public_bytes = public_key.public_bytes(
    encoding=serialization.Encoding.X962,
    format=serialization.PublicFormat.CompressedPoint
)

# Hash the message
dct_to_send = { '9': 9, '2': 2, '3': 3, '10': { '2': 2, '100': 100, '1': 1 } }
message = json.dumps(dct_to_send, sort_keys=True, separators=(',',':')).encode()
obj_hash = hashlib.sha256(message)
str_hash = obj_hash.hexdigest()
b_hash = obj_hash.digest()

# Sign the hash
signature_bytes = private_key.sign(b_hash, ec.ECDSA(hashes.SHA256()))

# Base64url encode the signature
signature_b64 = base64.urlsafe_b64encode(signature_bytes).decode()

pk_b64 = base64.urlsafe_b64encode(public_bytes).decode()

dct_msg = {
  **dct_to_send,
  'EE_SENDER': "0xai_" + pk_b64,
  'EE_SIGN': signature_b64,
  'EE_HASH': str_hash
}

str_msg = json.dumps(dct_msg) # no need for sorting
print("Message: '"+ str_msg + "'")






import os
import sys

from datetime import datetime as dt

class _TF2ModulesMixin(object):
  """
  Mixin for TF v2 modules functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_TF2ModulesMixin, self).__init__()

  def delete_module(self, label):
    return self.delete_tf2_module(label)

  def delete_tf2_module(self, label):
    import shutil
    save_folder = self.get_models_folder()
    module_folder = os.path.join(save_folder, label)
    if os.path.isdir(module_folder):
      shutil.rmtree(module_folder, ignore_errors=True)
      self.P("Remove {}".format(module_folder))
    return

  def save_module(self, module, label, unique_name=False, verbose=True):
    """
    Saves a tf.Module - wrapper for SaveTF2Module
    """
    return self.save_tf2_module(module=module, label=label, unique_name=unique_name, verbose=verbose)

  def save_tf2_module(self, module, label, unique_name=False, verbose=True):
    """
    Saves a tf.Module object

    Parameters
    ----------
    module : tf.Module
    label :  name of the model to be saved in `_models`
    unique_name : True if timestamp prefixing is required
    verbose : show info
    Returns
    -------
    None.

    """
    import tensorflow as tf
    save_folder = self.get_models_folder()
    file_prefix = ''
    if unique_name:
      file_prefix = dt.now().strftime("%Y%m%d_%H%M%S_")
    file_name = os.path.join(save_folder, file_prefix + label)
    # now stupid tf2 bug:
    if os.name == 'nt':
      file_name = file_name.replace('/', '\\')
    if verbose:
      self.verbose_log("Saving tf2 module [...{}]".format(file_name[-40:]))
    try:  # must try-except ... more tf2 bugs when saving
      tf.saved_model.save(module, file_name)
    except:
      _err = sys.exc_info()[0]
      self.P("ERROR: tf.saved_model.save failed: {}".format(_err))
    return

  def load_module(self, label, verbose=True):
    """
      Loads a tf.Module object for inference. Wrapper for LoadTF2Module
    """
    return self.load_tf2_module(label=label, verbose=verbose)

  def load_tf2_module(self, label, verbose=True):
    """
    Loads a tf.Module object for inference

    Parameters
    ----------
    label : name of the model from `_models` folder
    verbose : show info

    Returns
    -------
    mdl : tf.Module
    """
    import tensorflow as tf
    save_folder = self.get_models_folder()
    file_name = os.path.join(save_folder, label)
    mdl = tf.saved_model.load(file_name)
    self.P("Loaded tf2 module [...{}]".format(file_name[-40:]))
    return mdl



#global dependencies
import os
import cv2
import numpy as np
import pandas as pd

from time import time

#local dependencies
from naeural_core import constants as ct

DUMP_TIME_SECS = 60 * 1

### TODO: TO BE DELETED
class _ExecutionAuditMixin(object):
  def __init__(self):
    self._audit_fld = None
    self._data = self._empty_data()
    self._last_dump_time = time()
    super(_ExecutionAuditMixin, self).__init__()
    return


  @property
  def cfg_audit_dump_time(self):
    return self._instance_config.get('AUDIT_DUMP_SECS', DUMP_TIME_SECS)

  def _empty_data(self):
    dct = {
      'TIMESTAMP' : [],
      'TYPE'      : [],
      'MSG'       : [],
    }
    return dct

  def _maybe_setup_audit_fld(self):
    if self._audit_fld is not None:
      return
    stream_name, signature, instance_id = self.unique_identification
    self._audit_fld = os.path.join(ct.AUDIT_PLUGINS, signature, stream_name, instance_id)
    self.log.check_folder_output(self._audit_fld)
    return

  def _dump_audit_data(self):
    self._maybe_setup_audit_fld()
    df = pd.DataFrame(self._data)
    self.log.save_dataframe(
      df,
      fn='{}.csv'.format(self.log.now_str()),
      folder='output',
      subfolder_path=os.path.join(self._audit_fld, self.log.file_prefix),
      verbose=False
    )

    self._last_dump_time = time()
    self._data = self._empty_data()
    return

  def _maybe_dump_audit_data(self):
    if time() - self._last_dump_time < DUMP_TIME_SECS:
      return

    self._dump_audit_data()
    return

  def _handle_images(self, img):
    if img is None:
      return

    lst_img = img
    if not isinstance(img, list):
      lst_img = [img]

    #check all images are type np.ndarray
    is_ok = all([isinstance(x, np.ndarray) for x in lst_img])
    if not is_ok:
      self.P('Please be sure to provide np.ndarray objects in order to save them while audit your plugin execution!', color='y')
      return

    self._maybe_setup_audit_fld()
    names = []
    fld = self.log.check_folder_output(os.path.join(self._audit_fld, self.log.file_prefix, 'imgs'))
    for img in lst_img:
      if img.dtype != np.uint8:
        if img.dtype == 'float' and 0 <= img.min() <= 1 and 0 <= img.max() <= 1:
          img = img * 255.0
        # endif
        img = img.astype(np.uint8)
      #endif
      name = '{}.png'.format(self.log.now_str())
      names.append(name)
      cv2.imwrite(os.path.join(fld, name), img)
    #endfor
    return names

  def _handle_custom_information(self, **kwargs):
    #check if custom information was added in the message
    if len(kwargs) == 0:
      return

    for k, v in kwargs.items():
      if k not in self._data:
        self._data[k] = [''] * (len(self._data['TIMESTAMP']) - 1)
      self._data[k].append(v)
    #endfor
    return

  def _handle_data_fill(self):
    #check if all columns have the same number of records
    nr_max = max(len(v) for v in self._data.values())
    cols_to_fill = [k for k,v in self._data.items() if len(v) < nr_max]
    for col in cols_to_fill:
      self._data[col].append('')
    #endfor
    return

  def _add_message(self, msg, msg_type, img=None, **kwargs):
    if not self.cfg_enable_audit:
      return

    self._data['TIMESTAMP'].append(self.log.now_str())
    self._data['TYPE'].append(msg_type)
    self._data['MSG'].append(msg)

    names = self._handle_images(img)
    if names is not None:
      kwargs['IMG'] = names

    self._handle_custom_information(**kwargs)
    self._handle_data_fill()

    self._maybe_dump_audit_data()
    return

  def audit_log(self, msg, img=None, **kwargs):
    self._add_message(msg=msg, msg_type='log', img=img, **kwargs)
    return

  def audit_warning(self, msg, img=None, **kwargs):
    self._add_message(msg=msg, msg_type='warning', img=img, **kwargs)
    return

  def audit_error(self, msg, img=None, **kwargs):
    self._add_message(msg=msg, msg_type='error', img=img, **kwargs)
    return

  def audit_dump_audit_data(self):
    self._dump_audit_data()
    return


if __name__ == '__main__':
  from naeural_core import Logger
  from naeural_core import DecentrAIObject

  log = Logger(lib_name='SER', base_folder='.', app_folder='_local_cache', TF_KERAS=False)

  class Base(DecentrAIObject, _ExecutionAuditMixin):
    def __init__(self, **kwargs):
      self._instance_config = None
      super(Base, self).__init__(**kwargs)

    @property
    def unique_identification(self):
      return self._instance_config['STREAM'], self._instance_config['SIGNATURE'], self._instance_config['INSTANCE_ID']

    def startup(self):
      super().startup()
      self._instance_config = {
        'STREAM': '<STREAM_NAME>',
        'SIGNATURE': '<PLUGIN_SIGNATURE>',
        'INSTANCE_ID': '<PLUGIN_INSTANCE_ID>',
        'ENABLE_AUDIT': True
      }
      self.P("In base startup")
      return


  class Plugin(Base):
    def __init__(self, **kwargs):
      super(Plugin, self).__init__(**kwargs)
      return

    def startup(self):
      super().startup()
      return


  p = Plugin(log=log)
  p.audit_log('Test 1')
  p.audit_warning('Test 2')
  p.audit_error('Test 3')
  p.audit_log('Test4', a=1, b=2)
  p.audit_warning('Test5')
  p.audit_log('Test6', a=1, b=2, abcde='sdfsdf', kkkkkkk=12312312)
  p.audit_log('Test7', c=4, d=4, e=5)
  p.audit_log('Test8')
  p.audit_log('Test9', img=np.random.randint(0, 255, (400, 400, 3), dtype=np.uint8))
  p.audit_warning('Test10')
  p.audit_dump_audit_data()


MAX_LEN = 1000
IDS_DEQUE = 'IDS_DEQUE'
SEND_LOWERED = 'SEND_LOWERED'
IDS_SET = 'IDS_SET'


class _AlertTrackerMixin(object):
  """
  The purpose of this mixin is for us to determine if we should send
  a certain alert or not.
  To do that we will check if all the alertable entities in the current frame
  already created an alert before.
  If at least one entity E from the current frame was never in an alertable state
  when an alert was risen or if enough entities were seen between the last alert
  entity E was seen into and the current alert we will send the current event.
  The entity identification is based on a centroid object tracker.
  Also, 2 different Alerter objects will perform the above described analysis
  independently.
  """
  def __init__(self):
    """
    For each alerter we will keep track of its already alerted ids and whether
    or not we already sent a "raised alert" event and have to also send a "lowered
    alert" event.
    Each alerter will have its own dictionary with the following items:
      1) "IDS_DEQUE" - deque of the already alerted ids
      2) "IDS_SET" - set of already alerted ids (this is used strictly for efficiency reasons)
    """
    self.alerters_data = {}
    # self.alerted_ids = {}
    # self.send_lowered_alert = True
    # self.alerted_ids_set = {}
    super(_AlertTrackerMixin, self).__init__()
    return

  @property
  def cfg_alert_tracker_maxlen(self):
    return self._instance_config.get('ALERT_TRACKER_MAX_LEN', MAX_LEN)

  @property
  def cfg_send_all_alerts(self):
    return self._instance_config.get("SEND_ALL_ALERTS", False)

  def create_alerter_data(self, alerter='default'):
    self.alerters_data[alerter] = {
      IDS_DEQUE: self.deque(maxlen=self.cfg_alert_tracker_maxlen),
      IDS_SET: set(),
      SEND_LOWERED: True
    }
    return

  def full_deque(self, alerter='default'):
    ids_deque = self.alerters_data[alerter][IDS_DEQUE]
    return len(ids_deque) == ids_deque.maxlen

  def search_id(self, id, alerter='default'):
    if alerter not in self.alerters_data.keys():
      self.create_alerter_data(alerter=alerter)
    return id in self.alerters_data[alerter][IDS_SET]

  def add_track_id(self, id, alerter='default'):
    if not self.search_id(id=id, alerter=alerter):
      if self.full_deque(alerter=alerter):
        # in case the deque is full the first id in the deque will be removed from it
        # therefore, we have to remove it from the set too
        first_id, _ = self.alerters_data[IDS_DEQUE][0]
        self.alerters_data[alerter][IDS_SET].remove(first_id)
      now = self.time()
      self.alerters_data[alerter][IDS_DEQUE].append((id, now))
      self.alerters_data[alerter][IDS_SET].add(id)
      return True

    return False

  def add_track_ids(self, ids, alerter='default'):
    # returns True if at least one id is not alerted before
    # False otherwise
    if len(ids) > 0:
      return max(self.add_track_id(id=x, alerter=alerter) for x in ids)
    return False

  def check_event_sending(self, alert_ids, alert_state, alerter='default'):
    if self.cfg_send_all_alerts:
      return True

    if alerter not in self.alerters_data.keys():
      self.create_alerter_data(alerter=alerter)
    if alert_state:
      # raised alert
      # if the current alert is raised we check to see if we will send it
      self.alerters_data[alerter][SEND_LOWERED] = self.add_track_ids(ids=alert_ids, alerter=alerter)

    # in case the current alert is lowered we will send it only if we also sent the last raised alert
    return self.alerters_data[alerter][SEND_LOWERED]



"""
This module contains the BasicThAnchor class, which is a base class for comparing images to anchors.
The class is designed to hide as much as possible the lifecycle of the anchors in the serving process.
This class assumes that models have a 2 stage inference process (create embeddings and compare embeddings)
and that the models have the following methods exposed:
- get_embedding(th_image) -> th_embedding
- forward(th_embedding, th_image) -> th_prediction
"""

from naeural_core.serving.base import UnifiedFirstStage as BaseServingProcess
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

_CONFIG = {
  **BaseServingProcess.CONFIG,

  'IMAGE_HW': (480, 850),  # values should be taken from the training grid search

  "MODEL_WEIGHTS_FILENAME": None,
  "URL": None,
  'MAX_BATCH_FIRST_STAGE': 10,


  'ANCHOR_RELOAD_PERIOD': 24 * 60 * 60,  # seconds
  'ANCHOR_CHECK_NEWER_PERIOD': 1 * 60,  # seconds
  'ANCHOR_MAX_OLD_TIME': 3 * 24 * 60 * 60,  # seconds
  'DATE_FORMAT': '%Y/%m/%d_%H:%M:%S',
  'DEFAULT_DATE': '1000/10/10_00:00',

  "LOAD_PREVIOUS_SERIALIZATION": True,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

__VER__ = '0.3.0.0'


class BasicThAnchor(BaseServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._anchor_images = {}

    self._empty_anchor_embedding = None
    self._last_verbse_load_time = 0
    self._last_check_newer_anchor_time = 0
    self._verbose_load = True
    super(BasicThAnchor, self).__init__(**kwargs)
    return

  def _startup(self):
    msg = super()._startup()
    self._maybe_load_anchors_startup()

    self._empty_image = self.th.zeros(1, 3, *self.cfg_image_hw, dtype=self.th_dtype, device=self.dev)
    self._empty_anchor_embedding = self._get_anchor_embedding(self._empty_image)

    return msg

  def _get_model(self, config):
    """
    This method assumes the usage of torchscript models
    """
    model = self._prepare_ts_model(fn_model=config, post_process_classes=False)
    return model

  def model_call(self, th_inputs, model):
    """
    This method will perform a model inference on the given input and is used only during warmup.
    Assume the input is random.

    Parameters
    ----------
    th_inputs : torch.Tensor
        Input of the model. It will come as a single input, in this case as an image
    model : torch.nn.Module
        The model for which we perform the warmup

    Returns
    -------
    torch.Tensor
        The output of the model
    """
    th_anchor_embedding = model.get_embedding(th_inputs)
    return model(th_anchor_embedding, th_inputs)

  def _get_anchor_embedding(self, th_anchor_image):
    """
    Wrapper for the model's get_embedding method.
    Overwrite this method if you want to change the way the embedding is computed.
    """
    return self.th_model.get_embedding(th_anchor_image)

  def _save_anchor(self, unique_instance_identifier, anchor, update_time, metadata):
    """
    Save the anchor for the given instance.

    Parameters
    ----------
    unique_instance_identifier : str
        The unique instance identifier for the anchor.
    anchor : th.Tensor
        The anchor image.
    update_time : str
        The save time of the anchor.
    metadata : dict
        The metadata of the anchor.
    """
    th_transformed_image = self._transform_image(anchor, metadata)
    self._anchor_images[unique_instance_identifier] = {
      'image': th_transformed_image,
      'encoding': self._get_anchor_embedding(th_transformed_image),
      'update_time': update_time,
      'metadata': metadata
    }
    return

  def _save_known_anchor(self, known_anchor_images, unique_instance_identifier):
    """
    Save the known anchor for the given instance.
    Extracts the anchor from the known_anchor_images and saves it to the current anchor_images.

    Parameters
    ----------
    known_anchor_images : dict
        The dictionary containing the known anchor images from previous iteration.
    unique_instance_identifier : str
        The unique instance identifier for the anchor.
    """
    cached_entry = known_anchor_images[unique_instance_identifier]

    self._anchor_images[unique_instance_identifier] = {
      'image': cached_entry['image'],
      'encoding': cached_entry['encoding'],
      'update_time': cached_entry['update_time'],
      'metadata': cached_entry['metadata']
    }
    return

  def _maybe_load_anchors_startup(self):
    """
    This method is called in startup and loads the anchors from the previous serialization.
    """
    if self.cfg_load_previous_serialization:
      saved_data = self.persistence_serialization_load(default={})
      self._maybe_reset_anchors(saved_data)
    # endif
    return

  def _anchor_expired(self, unique_instance_identifier):
    """
    This method checks if the anchor is old enough to be changed.

    Parameters
    ----------
    unique_instance_identifier : str
        The unique instance identifier for which we check the anchor.

    Returns
    -------
    bool
        True if the anchor is old enough to be changed, False otherwise.
    """
    if unique_instance_identifier not in self._anchor_images:
      return True

    # if the reload period is None, we never reload the anchor
    if self.cfg_anchor_reload_period is None:
      return False

    last_recorded_update_time = self.datetime.strptime(
      self._anchor_images[unique_instance_identifier]['update_time'], self.cfg_date_format)
    update_time = self.datetime.now()
    seconds_since_last_update = (update_time - last_recorded_update_time).seconds
    return seconds_since_last_update >= self.cfg_anchor_reload_period

  def _new_anchor_available(self, known_anchor_images, unique_instance_identifier, update_time):
    """
    This method checks if a new anchor is available for an instance.

    Parameters
    ----------
    known_anchor_images: dict
        The dictionary containing the known anchor images from previous iteration.
    unique_instance_identifier : str
        The unique instance identifier for which we check the anchor.
    update_time : datetime
        The time of the new anchor.

    Returns
    -------
    bool
        True if a new anchor is available, False otherwise.
    """
    # if the anchor is not in known, we load it
    if unique_instance_identifier not in known_anchor_images:
      return True

    last_recorded_update_time = self.datetime.strptime(
      known_anchor_images[unique_instance_identifier]['update_time'], self.cfg_date_format)

    # if the anchor is newer than the last recorded update time, we load it
    if (update_time - last_recorded_update_time).seconds > 0:
      return True

    return False


  def _ready_to_check_newer_anchor(self):
    """
    This method checks if it is time to check for newer anchors.

    Returns
    -------
    bool
        True if it is time to check for newer anchors, False otherwise.
    """
    current_time = self.time()
    seconds_since_last_check = (current_time - self._last_check_newer_anchor_time)
    return seconds_since_last_check > self.cfg_anchor_check_newer_period

  def _maybe_reset_anchors(self, saved_data):
    """
    This method checks if there are newer anchors available and resets the anchors if needed.

    Parameters
    ----------
    saved_data : dict
        The saved data from the previous serialization. (The data loaded from disk)
    """

    known_anchor_images = self._anchor_images
    self._anchor_images = {}

    # this method now resets only changes, not all images
    dct_pipeline_anchors_images = saved_data.get(self.ct.ThAnchor.ANCHOR_IMAGES, {})
    dct_pipeline_anchor_metadata = saved_data.get(self.ct.ThAnchor.ANCHOR_METADATA, {})

    for unique_instance_identifier, pipeline_metadata in dct_pipeline_anchor_metadata.items():
      # anchor image does not exist, discard entry
      if unique_instance_identifier not in dct_pipeline_anchors_images:
        continue

      # anchor is too old, discard entry
      str_pipeline_last_update_time = pipeline_metadata.get(self.ct.ThAnchor.ANCHOR_UPDATED_AT, self.cfg_default_date)
      pipeline_last_update_time = self.datetime.strptime(str_pipeline_last_update_time, self.cfg_date_format)
      if self.cfg_anchor_max_old_time is not None and (self.datetime.now() - pipeline_last_update_time).seconds > self.cfg_anchor_max_old_time:
        continue

      # anchor is new, compute embedding and save
      if self._new_anchor_available(known_anchor_images, unique_instance_identifier, pipeline_last_update_time):
        self._save_anchor(
          unique_instance_identifier=unique_instance_identifier,
          anchor=dct_pipeline_anchors_images[unique_instance_identifier],
          update_time=str_pipeline_last_update_time,
          metadata=pipeline_metadata
        )
      # anchor is not new but is known, load from cache
      else:
        self._save_known_anchor(known_anchor_images, unique_instance_identifier)
        
    return

  def _comment_reason_for_spam(self, verbose):
    """
    This method prints a warning if there are no anchors available for any pipeline.

    Parameters
    ----------
    verbose : bool
        True if the warning should be printed, False otherwise.
    """
    bad_pipelines = [
        x for x in self._anchor_images.keys()
        if self._anchor_expired(x)
      ]

    if verbose:
      if len(self._anchor_images) == 0:
        self.P("Warning! No anchors found for any pipeline! This is an expected behavior if the pipeline has just been started...", color='r')

      if len(bad_pipelines) > 0:
        self.P("Warning! Unable to retrieve newer anchors for pipelines {}. Please see plugin warnings for detailed information.".format(
          str(bad_pipelines)), color='r')

    return

  def get_pipeline_name(self, uiid):
    """
    This method extracts the pipeline name from the unique instance identifier.

    Parameters
    ----------
    uiid : str
        The unique instance identifier.

    Returns
    -------
    str
        The pipeline name.
    """
    # unique_instance_identifier = '(pipeline_name, plugin_signature, instance_id)'
    return uiid.split('\'')[1]

  def _check_anchor_update(self, pipeline_names):
    """
    This method checks if the anchors should be updated and updates them if needed.
    """
    can_reset_anchors = (
      len(self._anchor_images) == 0
      or any([self._anchor_expired(x) for x in self._anchor_images.keys()])
      or self._ready_to_check_newer_anchor()
      or any(len([uiid for uiid in self._anchor_images if self.get_pipeline_name(uiid) == pipeline_name]) == 0 for pipeline_name in pipeline_names)
    )
    if can_reset_anchors:
      # possible fail here, should turn off verbosity
      if self.time() - self._last_verbse_load_time > 60:  # seconds
        self._verbose_load = True
        self._last_verbse_load_time = self.time()
      else:
        self._verbose_load = False

      self._comment_reason_for_spam(verbose=self._verbose_load)
      saved_data = self.persistence_serialization_load(default={}, verbose=self._verbose_load)
      self._maybe_reset_anchors(saved_data)
      self._last_check_newer_anchor_time = self.time()
    return

  def _transform_image(self, image, metadata):
    """
    This method transforms the image to the desired format.
    This method outputs an image in RGB format, normalized to [-1, 1].

    Parameters
    ----------
    image : th.Tensor
        The image to be transformed.

    Returns
    -------
    th.Tensor
        The transformed image.
    """
    image, _ = th_resize_with_pad(
      img=image,
      h=self.cfg_image_hw[0],
      w=self.cfg_image_hw[1],
      normalize=True,
      sub_val=127.5,
      div_val=127.5,
      device=self.dev,
      half=self.cfg_fp16
    )

    image = image.unsqueeze(0)
    return image

  def _pre_process(self, inputs):
    """
    Preprocess the inputs and create the inputs for the model.

    Parameters
    ----------
    inputs : dict
        The inputs to be preprocessed.

    Returns
    -------
    dict
        The preprocessed inputs.
    """
    input_images = inputs[self.ct.DATA]
    # TODO: maybe change this key from STREAM_NAME to PIPELINE_NAME
    pipeline_names = inputs[self.ct.PAYLOAD_DATA.STREAM_NAME]

    # check if we need to update the anchors
    self._check_anchor_update(pipeline_names)

    # save a mapping from pipeline name to image
    dct_pipeline_name_image = {pipeline_name: input_image for pipeline_name, input_image in zip(pipeline_names, input_images)}

    input_images = []
    anchors = []
    anchor_load_times = []

    for unique_instance_identifier in self._anchor_images:
      pipeline_name = self.get_pipeline_name(unique_instance_identifier)

      if pipeline_name not in dct_pipeline_name_image:
        # did not receive an input image from this pipeline
        input_images.append(self._empty_image)
        anchors.append(self._empty_anchor_embedding)
        anchor_load_times.append(None)
      else:
        # we have an input image from this pipeline
        metadata = self._anchor_images[unique_instance_identifier]['metadata']
        pipeline_image = dct_pipeline_name_image[pipeline_name]

        input_images.append(self._transform_image(pipeline_image, metadata))
        anchors.append(self._anchor_images[unique_instance_identifier]['encoding'])
        anchor_load_times.append(self._anchor_images[unique_instance_identifier]['update_time'])
      # endif pipeline_name in dct_pipeline_name_image
    # endfor prepare test_images

    if len(input_images) == 0:
      # if we did not receive input from any pipeline, we will process an empty image
      input_images.append(self._empty_image)
      anchors.append(self._empty_anchor_embedding)
    # endif

    th_images = self.th.cat(input_images, dim=0)
    th_anchor_encodings = self.th.cat(anchors, dim=0)

    self._anchor_load_times = anchor_load_times
    self._pipeline_names = pipeline_names

    prep_inputs = {
      'inputs': {
        'th_image': th_images,
        'th_anchor_embedding': th_anchor_encodings
      }

    }

    return prep_inputs

  def _get_inputs_batch_size(self, inputs):
    """
    We overwrite this method since we have two inputs, not one
    """
    return len(inputs['th_image'])

  def _post_process(self, preds):
    """
    Postprocess the predictions and create the outputs for the model.
    This method can be overwritten to change the way the outputs are created.

    Parameters
    ----------
    preds : th.Tensor
        The predictions from the model.

    Returns
    -------
    list
        The postprocessed outputs.
    """
    preds = preds.detach().cpu().numpy()

    # uiid = unique_instance_identifier
    dct_uiid_pred_index = {uiid: j for j, uiid in enumerate(self._anchor_images.keys())}

    lst = []
    for pipeline_name in self._pipeline_names:
      dct_uiid_preds = {
        self.ct.ThAnchor.ANCHOR_RELOAD_TIME: self.cfg_anchor_reload_period,
      }

      for uiid in self._anchor_images:
        if self.get_pipeline_name(uiid) == pipeline_name:
          dct_uiid_preds[uiid] = {
            self.ct.ThAnchor.PREDICTION: preds[dct_uiid_pred_index[uiid]],
            self.ct.ThAnchor.ANCHOR_UPDATED_AT: self._anchor_load_times[dct_uiid_pred_index[uiid]],
          }
        # endif
      # end for uiid
      lst.append([dct_uiid_preds])
    # end for pipeline_name

    return lst



import os
import sys

class _DeployModelsInProductionMixin(object):
  """
  Mixin for deployable models functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_DeployModelsInProductionMixin, self).__init__()

  def save_code(self, obj, name, folder):
    import cloudpickle
    info = sys.version_info
    sver = '_{}{}{}'.format(info.major, info.minor, info.micro)
    assert not any(char.isdigit() for char in name), "name must not contain numbers - {}".format(name)

    fn = os.path.join(folder, name + sver + '.dat')
    with open(fn, 'wb') as fh:
      cloudpickle.dump(obj, fh)
    self.P("Saved '{}'".format(fn))
    return

  def load_code(self, name, folder):
    import pickle
    import cloudpickle
    vmm = int(cloudpickle.__version__[:3].replace('.', ''))
    if vmm < 16:
      self.raise_error("You must have cloudpickle >= 1.6.0. Found {}".format(
        cloudpickle.__version__))
    info = sys.version_info
    sver = '_{}{}{}'.format(info.major, info.minor, info.micro)
    assert not any(char.isdigit() for char in name), "name must not contain numbers - {}".format(name)

    fn = os.path.join(folder, name + sver + '.dat')

    if not os.path.isfile(fn):
      fnhs = [x for x in os.listdir(folder) if name in x]
      raise ValueError('File {} not found. Avail files are: {}'.format(
        fn, fnhs))

    with open(fn, 'rb') as fh:
      data = pickle.load(fh)
    self.P("Loaded '{}'".format(fn))
    return data

  def _load_helper(self):
    helper = self.load_code('model_helper','libraries')
    return helper

  def load_deployed_library(self, filename, debug=False, **kwargs):
    """
    Loads a deployed library/module (.lib)

    Parameters
    ----------
    filename : str
      relative path of the file including the filename.

    Returns
    -------
    reference to the module.

    """
    _f = os.path.splitext(filename)[0]
    ext = os.path.splitext(filename)[-1].lower()
    ext = ''.join([chr(x) for x in [46, 112, 121]]) if ext == '' else ext
    filename = _f + ext
    if not os.path.isfile(filename):
      filename = _f + '.lib'
      if not os.path.isfile(filename):
        raise ValueError("File '{}' not found!".format(filename))
      else:
        ext = '.lib'
    if ext != '.lib':
      if ext[-1] == 'y':
        import importlib
        module_name = filename.replace('/','.')[:-3]
        module_name = module_name.replace('\\','.')
        return importlib.import_module(module_name)
      else:
        raise ValueError("Deployed libraries must have '.lib' extension. Received: '{}'".format(
          filename))
    helper = self._load_helper()
    lib = helper.load_lib(self, filename, DEBUG=debug, **kwargs)
    return lib

  def save_deployable_library(self, filename, dest_folder=None, EXT='.lib'):
    """
    Saves a '.py' library/module in deploy format (.lib)

    Parameters
    ----------
    filename : str
      relative path of the file including the filename.

    dest_folder: str, defalut=None
      destination folder for the deployable library

    EXT: str, default='.lib'
      default deployable library extension


    Returns
    -------
    reference to the module.

    """
    EXT = '.lib'
    if not os.path.isfile(filename):
      raise ValueError("File '{}' not found!".format(filename))
    fn = os.path.splitext(filename)
    if fn[-1].lower() != '.py':
      raise ValueError("Source libraries must have '.py' extension. Received: '{}'".format(
        filename))
    src = filename
    if dest_folder is not None:
      fn_only = os.path.splitext(os.path.split(filename)[1])[0]
      dst = os.path.join(dest_folder, fn_only + EXT)
    else:
      dst = fn[0] + EXT
    self.P("Saving '{}'".format(dst))
    helper = self._load_helper()
    helper.save_lib(
      self,
      src_lib=src,
      dst_lib=dst
    )
    return

  def save_deploy_model(self, model, name, where='models'):
    """
    Saves a model for deployment

    Parameters
    ----------
    model : tf model
      the model object.
    name : str
      name of the model or full path to the output file
    where : str, optional
      'models' if `name` is a simple name or `fullpath` if full path is provided.
      The default is 'models'.

    Returns
    -------
    None.

    """
    assert where in ['models', 'fullpath']

    helper = self._load_helper()
    assert helper is not None

    if name[-3:] == '.h5':
      name = name[:-3]

    if where == 'models':
      model_file = os.path.join(self.get_models_folder(), name)
    else:
      model_file = name

    model_file += '.dat'

    helper.save_for_serving(
      log=self,
      model=model,
      deploy_model_path=model_file
    )
    return

  def load_deploy_model(self,
                        name,
                        where='models',
                        custom_objects=None,
                        DEBUG=False):
    """
    Loads a deployed model in production

    Parameters
    ----------
    name : str
      name (if in 'models' folder) or full path to the model
    where : str, optional
      either 'models' or 'fullpath'. The default is 'models'.

    Returns
    -------
    tf model

    """
    assert where in ['models', 'fullpath']
    import tensorflow as tf

    helper = self._load_helper()
    assert helper is not None

    if where == 'models':
      model_file = os.path.join(self.get_models_folder(), name)
    else:
      model_file = name

    if not os.path.isfile(model_file):
      raise ValueError("Model data not found '{}'".format(model_file))

    helper.load_for_serving(
      log=self,
      deploy_model_path=model_file,
      load_func=tf.keras.models.load_model,
      custom_objects=custom_objects,
      DEBUG=DEBUG,
    )

    return helper



import requests
from datetime import datetime

from naeural_core.utils.sensors.base import AbstractSensor

_SERVER = 'https://home.sensibo.com/api/v2'

class SensiboSensor(AbstractSensor):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return

  def _setup_connection(self, device_name, api_key, server_addr=_SERVER):
    self._device_name = device_name
    self._api_key = api_key
    self._server_addr = server_addr
    devices = self.__list_devices()
    self._uid = devices[device_name]
    return
  
  def _get_single_observation(self):
    obs = self.__get_last_measurement()
    return obs
    
  def _has_training_data(self):
    return None
    
  def _get_training_data(self):
    return False    
  
  def __get_last_measurement(self):
    res = self.__get_measurement(self._uid)
    return res[-1]  

  def __get_data(self, path, **params):
    params['apiKey'] = self._api_key
    response = requests.get(self._server_addr + path, params=params)
    response.raise_for_status()
    return response.json()

  def __patch(self, path, data, **params):
    params['apiKey'] = self._api_key
    response = requests.patch(self._server_addr + path, params=params, data=data)
    response.raise_for_status()
    return response.json()

  def __list_devices(self):
    result = self.__get_data('/users/me/pods', fields='id,room')
    return {x['room']['name']: x['id'] for x in result['result']}

  def __get_measurement(self, pod_uid=None):
    if pod_uid is None:
      pod_uid = self._uid
    results = self.__get_data('/pods/{}/measurements'.format(pod_uid))
    results = results['result']
    for res in results:
      if 'time' in res:
        str_dt = res['time']['time']
        dt = datetime.strptime(str_dt, '%Y-%m-%dT%H:%M:%S.%fZ')
        delay = res['time']['secondsAgo']
        res['read_time'] = dt
        res['read_time_str']  = str_dt
        res['read_delay'] = delay
        
    return results
    
  
  
if __name__ == "__main__":
  api_key = '0B073b470DeXHoqmXmdeBpVzBbHcLh'
  device_name = "Alex's device"
  
  sensor = SensiboSensor(api_key=api_key,device_name=device_name)
  print(sensor)
  print(sensor._get_single_observation())



import os
from pathlib import Path
import torch as th
from openvino.runtime import Core
import openvino as ov
import json
import numpy as np

from naeural_core.serving.base.backends.model_backend_wrapper import ModelBackendWrapper

from typing import Tuple

def get_torch_dtype(np_dtype : np.dtype) -> th.dtype:
  """
  Get a torch dtype from a numpy dtype.
  Parameters:
    - np_dtype a numpy dtype
  Returns:
    - the corresponding torch dtype.
  """
  numpy_to_torch_dtype_dict = {
      bool          : th.bool,
      np.bool_      : th.bool,
      np.uint8      : th.uint8,
      np.int8       : th.int8,
      np.int16      : th.int16,
      np.int32      : th.int32,
      np.int64      : th.int64,
      np.float16    : th.float16,
      np.float32    : th.float32,
      np.float64    : th.float64,
      np.complex64  : th.complex64,
      np.complex128 : th.complex128
  }
  if np_dtype not in numpy_to_torch_dtype_dict.keys():
    raise ValueError("No known corresponding torch dtype for {}".format(np_dtype))
  return numpy_to_torch_dtype_dict[np_dtype]

class OpenVINOModel(ModelBackendWrapper):
  ONNX_METADATA_KEY = 'onnx_metadata'

  def __init__(self):
    self._model = None
    self._compiled_model = None
    self._metadata = None
    return

  def load_model(
    self,
    model_path : str,
    half : bool
  ) -> None:
    """
    Loads the OpenVINO from disk (in ONNX format).
    If the model cannot be loaded, raises an Exception.

    Parameters
    ----------

    model_path : str
      The path of the model on disk, in ONNX format
    half : bool
      If True loads the model in fp16

    Returns
    -------
    None
    """
    import onnx
    from google.protobuf.json_format import MessageToDict
    # We're only going to run this on the CPU so hard-code it here.
    # We could in theory use other devices however the output seems
    # to always be on the CPU side (i.e. in a numpy array).
    device = 'CPU'

    # Read the metadata from the model.
    model_onnx = onnx.load(model_path)
    onnx_metadata_dict = MessageToDict(model_onnx, preserving_proto_field_name = True)['metadata_props']
    for meta in onnx_metadata_dict:
      if meta['key'] == OpenVINOModel.ONNX_METADATA_KEY:
        self._metadata = json.loads(meta['value'])

    model_precision = self._metadata.get('precision')
    if model_precision is not None:
      if (model_precision.lower() == "fp16") != half:
        if half and model_precision.lower() == "fp32":
          #from onnxconverter_common import float16
          from onnxruntime.transformers import float16
          model_fp16 = float16.convert_float_to_float16(model_onnx, disable_shape_infer=True)
          fp16path = str(Path(model_path).with_suffix('.tofp16.onnx'))
          onnx.save(model_fp16, fp16path)
          model_path = fp16path
        else:
          raise RuntimeError('Incompatible precision in ONNX model')
    #endif check onnx precision

    del model_onnx
    model_onnx = None

    # Building the OpenVINO model from ONNX is fast so there is no
    # issue with rebuilding on load.
    if not OpenVINOModel.create_from_onnx(onnx_path=model_path):
      raise RuntimeError("Error while creating OpenVINO model")

    core = Core()
    self._model = core.read_model(
      model=model_path,
      weights=Path(model_path).with_suffix('.bin')
    )
    self._compiled_model = core.compile_model(self._model, device_name=device)

    return

  def forward(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    args_dict = {}
    # Set model arguments while ensuring that we're using contiguous tensors
    # and the data is on the CPU.
    for i, _ in enumerate(self._model.inputs):
      args_dict[i] = args[i].cpu().contiguous().numpy()

    request = self._compiled_model.create_infer_request()
    result = request.infer(inputs=args_dict) # Run inference.

    # Convert output into a tuple of torch tensors. Note that we're
    # already running on the CPU so this shouldn't involve any
    # memory copies.
    ret = tuple([th.tensor(result[out]) for out in request.model_outputs])
    if len(ret) == 1:
      return ret[0]
    return ret

  def __call__(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    """
    Override for the call operator, just goes to forward.
    Input and output tensors are in the ONNX order.
    """
    return self.forward(*args)

  @staticmethod
  def create_from_onnx(
    onnx_path : str,
    vino_path : str = None,
    half = False
  ):
    status = True
    if vino_path is None:
      vino_path = str(Path(str(onnx_path)).with_suffix('.xml'))
    #endif

    weights_path = Path(vino_path).with_suffix('.bin')
    try:
      ov_model = ov.convert_model(onnx_path, share_weights=False)
      ov.serialize(ov_model, vino_path)
    except Exception as e:
      status = False

    if not status:
      # Something went wrong, remove all vino files (xml and bin files).
      try:
        os.unlink(vino_path)
      except Exception:
        pass
      try:
        os.unlink(weights_path)
      except Exception:
        pass

    return status

  def get_metadata(self):
    return self._metadata

  def get_device(self) -> th.device:
    return th.device('cpu')

  def get_input_dtype(self, index : int) -> th.dtype:
    np_dtype = self._model.inputs[index].get_element_type().to_dtype().type
    return get_torch_dtype(np_dtype)

  def __call__(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    return self.forward(*args)
